import torch
import torch.nn as nn
import torch.optim as optim
from alcokit.models.modules import ParamedSampler, Pass
from alcokit.models.losses import weighted_L1
from alcokit.models import Model


class EncoderLSTM(nn.Module):
    def __init__(self, input_d, model_dim, num_layers, bottleneck="add", n_fc=1):
        super(EncoderLSTM, self).__init__()
        self.bottleneck = bottleneck
        self.dim = model_dim
        self.lstm = nn.LSTM(input_d, self.dim if bottleneck == "add" else self.dim // 2,
                            num_layers=num_layers, batch_first=True, bidirectional=True)
        self.fc = nn.Sequential(
            *[nn.Sequential(nn.Linear(self.dim, self.dim), nn.Tanh()) for _ in range(n_fc - 1)],
            nn.Linear(self.dim, self.dim),  # NO ACTIVATION !
        )

    def forward(self, x, hiddens=None, cells=None):
        if hiddens is None or cells is None:
            states, (hiddens, cells) = self.lstm(x)
        else:
            states, (hiddens, cells) = self.lstm(x, (hiddens, cells))
        states = self.first_and_last_states(states)
        return self.fc(states), (hiddens, cells)

    def first_and_last_states(self, sequence):
        sequence = sequence.view(*sequence.size()[:-1], self.dim, 2).sum(dim=-1)
        first_states = sequence[:, 0, :]
        last_states = sequence[:, -1, :]
        if self.bottleneck == "add":
            return first_states + last_states
        else:
            return torch.cat((first_states, last_states), dim=-1)


class DecoderLSTM(nn.Module):
    def __init__(self, model_dim, num_layers, bottleneck="add"):
        super(DecoderLSTM, self).__init__()
        self.dim = model_dim
        self.lstm1 = nn.LSTM(self.dim, self.dim if bottleneck == "add" else self.dim // 2,
                             num_layers=num_layers, batch_first=True, bidirectional=True)
        self.lstm2 = nn.LSTM(self.dim, self.dim if bottleneck == "add" else self.dim // 2,
                             num_layers=num_layers, batch_first=True, bidirectional=True)

    def forward(self, x, hiddens, cells):
        if hiddens is None or cells is None:
            output, (_, _) = self.lstm1(x)
        else:
            output, (_, _) = self.lstm1(x, (hiddens, cells))  # V1 decoder DOES GET hidden states from enc !
        output = output.view(*output.size()[:-1], self.dim, 2).sum(dim=-1)

        if hiddens is None or cells is None:
            output2, (hiddens, cells) = self.lstm2(output)
        else:
            output2, (hiddens, cells) = self.lstm2(output, (hiddens, cells))  # V1 residual DOES GET hidden states from first lstm !
        output2 = output2.view(*output2.size()[:-1], self.dim, 2).sum(dim=-1)

        return output + output2, (hiddens, cells)


class Seq2SeqLSTM(nn.Module):
    def __init__(self, input_dim, model_dim,
                 num_layers=1,
                 bottleneck="add",
                 n_fc=1):
        super(Seq2SeqLSTM, self).__init__()
        self.enc = EncoderLSTM(input_dim, model_dim, num_layers, bottleneck, n_fc)
        self.dec = DecoderLSTM(model_dim, num_layers, bottleneck)
        self.sampler = ParamedSampler(model_dim, model_dim, pre_activation=Pass)

    def forward(self, x, output_length=None):
        coded, (h_enc, c_enc) = self.enc(x)
        if output_length is None:
            output_length = x.size(1)
        coded = coded.unsqueeze(1).repeat(1, output_length, 1)
        residuals, _, _ = self.sampler(coded)
        coded = coded + residuals
        output, (_, _) = self.dec(coded, h_enc, c_enc)
        return output



import unittest
from alcokit.sources import NPSource, TorchSource
import numpy as np
import torch
from alcokit.score import Score


class TestSources(unittest.TestCase):

    def test_np_persistent_type(self):
        arr = np.arange(36).reshape(6, 6)
        nps = NPSource(arr)

        assert isinstance(nps.flat[:], NPSource)

        assert isinstance(nps[0] + 1, NPSource)

        assert isinstance(nps[:, -1] * 8, NPSource)

    def test_torch_persistent_type(self):
        tsr = torch.arange(36).view(6, 6).float()
        ts = TorchSource(tsr)

        # TODO: Unfortunately not supported in pytorch yet...
        # assert isinstance(ts.view(-1), TorchSource)
        #
        # assert isinstance(ts[0] + 1, TorchSource)
        #
        # assert isinstance(ts[:, -1] * 8, TorchSource)

    def test_algos_on_arrays(self):
        from alcokit.alcokit.algos import harmonic_percussive, REP_SIM
        nps = NPSource(np.random.rand(256, 100))
        h, p = harmonic_percussive(nps)

        assert isinstance(h, np.ndarray)
        assert isinstance(p, np.ndarray)

        assert isinstance(h, NPSource)
        assert isinstance(p, NPSource)

        bg, fg = REP_SIM(nps)

        assert isinstance(bg, NPSource)
        assert isinstance(fg, NPSource)

    def test_scored_source_methods(self):
        ds = NPSource(np.arange(36).reshape(6, 6))

        assert isinstance(ds[0], NPSource)

        assert isinstance(ds[slice(0, 4)], NPSource)

        score = Score.from_duration([2, 2, 2])

        colcat = ds.colcat(score)
        assert isinstance(colcat, NPSource)
        assert colcat.shape[1] == 6

        uncat = ds.uncat(score)
        assert len(uncat) == 3

        colstack = ds.colstack(score)
        assert colstack.shape[0] == 3

        score["stop"] += 2
        patched = ds.patch(ds[:, -4:], (3, 3))
        assert patched.shape[1] == 7


if __name__ == '__main__':
    unittest.main()


from alcokit.util import normalize, m2hz
import numpy as np
import torch

# WIP


def spectra(S, midi_pitches,
            peaks="laplace",  # add ["sin", "saw", "sqr", ]
            h_dist=1., tuning=0., decay_len=1.,
            peak_width=.25, max_partial=4.5):
    """
    @param S:
    @param midi_pitches:
    @param peaks:
    @param h_dist: harmonic distorsion -> partial k = f0 * k * h_dist
    @param tuning: deviation in semi-tones to the standard tuning
    @param decay_len: ~ how many octaves it takes to kill the spectrum to 0. amplitude
    @param peak_width:
    @param max_partial:
    @return:
    """
    frequencies = np.linspace(0, 22050 / 2, S.shape[0])
    mid_col = m2hz(midi_pitches + tuning)
    ratios = (frequencies / mid_col[:, None]) ** (1/h_dist)
    half_col = .5 * mid_col[:, None]
    d = (((half_col + np.maximum(frequencies, half_col)) % mid_col[:, None]) - half_col)
    # normalize each peak by forcing regions near 0 to be even closer to 0
    eps = (frequencies[1] - frequencies[0]) / 2
    if peaks == "gauss":
        less_than_eps = np.sqrt(d ** 2) <= eps
        d[less_than_eps] = 0
        d = np.exp(-(1 / (peak_width ** 2)) * (d ** 2))
    elif peaks == "laplace":
        less_than_eps = abs(d) <= eps
        d[less_than_eps] = 0
        d = np.exp(-(1 / peak_width) * abs(d))

    # exponential decay of the spectrum
    d[(ratios > max_partial)] = 0
    d *= np.exp(- ratios * (1 / decay_len))
    d = normalize(d, axis=1)
    return d


def update_score(y, instrument, score, loss_fn, loss_t):
    with torch.no_grad():
        arange = torch.arange(score.size(1))
        temp = score + torch.zeros_like(score)
        mins = score.grad.min(dim=0)
        update = mins.values / torch.norm(score.grad, p=2, dim=0)
        temp[mins.indices, arange] -= update
        new_loss = loss_fn(y, torch.mm(instrument, temp))
        updatable = new_loss < loss_t
        score[mins.indices, arange][updatable] -= update[updatable]
        score.grad.zero_()
    return score


def optim_loop(X, loss_fn, n_iter):
    Y = abs(X)
    y_nz = Y.sum(axis=0) > 0
    instru = spectra(Y, np.arange(128),
                     "laplace", tuning=0.08, h_dist=1.015,
                     decay_len=np.linspace(2, 2, 128)[:, None],
                     peak_width=np.linspace(24, 24, 128)[:, None],
                     max_partial=22.5)

    T, P = Y.shape[1], instru.shape[0]

    Yt = torch.tensor(Y)
    instrument = torch.tensor(instru.T).float().requires_grad_(False)
    score = torch.zeros(P, T).float().requires_grad_(True)

    losses = []
    predictions = torch.mm(instrument, score)
    loss = loss_fn(Yt, predictions)
    L = loss.mean()
    L.backward(retain_graph=True)
    losses += [L.item()]
    for i in range(n_iter):
        print("iter:", i)
        score = update_score(Yt, instrument, score, loss_fn, loss)
        predictions = torch.mm(instrument, score)
        loss = loss_fn(Yt, predictions)
        L = loss.mean()
        L.backward(retain_graph=True)
        losses += [L.item()]



import h5py
import numpy as np
import pandas as pd
from alcokit.score import Score
from alcokit.sources import ScoredSourceMixin, NPSource
from functools import partial


class FeatureProxy(ScoredSourceMixin):

    time_axis = 0

    @property
    def _constructor(self):
        return NPSource

    @property
    def cat(self):
        # getitem transposes !!
        return lambda x: self._constructor(np.concatenate(x, axis=1))

    @property
    def stack(self):
        return np.stack

    def __init__(self, h5_file, ds_name, transposed=True):
        self.h5_file = h5_file
        self.name = ds_name
        self.T = transposed
        with h5py.File(h5_file, "r") as f:
            ds = f[self.name]
            self.N = ds.shape[0]
            self.attrs = {k: v for k, v in ds.attrs.items()}

    def __len__(self):
        return self.N

    def __getitem__(self, item):
        with h5py.File(self.h5_file, "r") as f:
            rv = f[self.name][item]
        return rv.T if self.T else rv


class Database(object):
    def __init__(self, h5_file):
        self.h5_file = h5_file
        self.info = self._get_dataframe("/info")
        self.score = Score(self._get_dataframe("/score"))
        with h5py.File(h5_file, "r") as f:
            # add found features as self.feature_name = FeatureProxy(self, feature_name, self.segs)
            self._register_features(self.features)

    @property
    def features(self):
        names = self.info.iloc[:, 2:].T.index.get_level_values(0)
        return list(set(names))

    def visit(self, func=print):
        with h5py.File(self.h5_file, "r") as f:
            f.visititems(func)

    def _get_dataframe(self, key):
        try:
            return pd.read_hdf(self.h5_file, key=key)
        except KeyError:
            return pd.DataFrame()

    def save_dataframe(self, key, df):
        with h5py.File(self.h5_file, "r+") as f:
            if key in f:
                f.pop(key)
        df.to_hdf(self.h5_file, key=key, mode="r+")
        return self._get_dataframe(key)

    def layout_for(self, feature):
        with h5py.File(self.h5_file, "r") as f:
            if "layouts" not in f.keys():
                return pd.DataFrame()
        return self._get_dataframe("layouts/" + feature)

    def _register_features(self, names):
        for name in names:
            setattr(self, name, FeatureProxy(self.h5_file, name))
        return None






import h5py
import numpy as np
import pandas as pd
import os
from multiprocessing import cpu_count, Pool
from alcokit.util import fs_dict, is_audio_file
from alcokit.hdf.api import Database
from alcokit.score import Score
import logging

logger = logging.getLogger()
logger.setLevel(logging.INFO)


# TODO : add handling of sparse matrices ?

def default_extract_func(abs_path):
    from alcokit.fft import FFT
    fft = abs(FFT.stft(abs_path))
    score = Score.from_recurrence_matrix(fft)
    return dict(fft=({}, fft.T), score=({}, score))


def sizeof_fmt(num, suffix='b'):
    """
    straight from https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
    """
    for unit in ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def _empty_info(features_names):
    tuples = [("directory", ""), ("name", ""),
              *[t for feat in features_names for t in [(feat, "dtype"), (feat, "shape"), (feat, "size")]
                if feat != "score"]
              ]
    idx = pd.MultiIndex.from_tuples(tuples)
    return pd.DataFrame([], columns=idx)


def split_path(path):
    parts = path.split("/")
    prefix, file_name = "/".join(parts[:-1]), parts[-1]
    return prefix, file_name


def file_to_db(abs_path, extract_func=default_extract_func, mode="w"):
    """
    if mode == "r+" this will either:
        - raise an Exception if the feature already exists
        - concatenate data along the "feature_axis", assuming that each feature correspond to the same file
          or file collections.
          If you want to concatenate dbs along the "file_axis" consider using `concatenate_dbs(..)`
    @param abs_path:
    @param extract_func:
    @param mode:
    @return:
    """
    logger.info("making db for %s" % abs_path)
    tmp_db = ".".join(abs_path.split(".")[:-1] + ["h5"])
    rv = extract_func(abs_path)
    info = _empty_info(rv.keys())
    info.loc[0, [("directory", ""), ("name", "")]] = split_path(abs_path)
    with h5py.File(tmp_db, mode) as f:
        for name, (attrs, data) in rv.items():
            if issubclass(type(data), np.ndarray):
                ds = f.create_dataset(name=name, shape=data.shape, data=data)
                ds.attrs.update(attrs)
                info.loc[0, name] = ds.dtype, ds.shape, sizeof_fmt(data.nbytes)
            elif issubclass(type(data), pd.DataFrame):
                pd.DataFrame(data).to_hdf(tmp_db, name, "r+")
            f.flush()
        if "info" in f.keys():
            prior = pd.read_hdf(tmp_db, "info", "r")
            info = pd.concat((prior, info.iloc[:, 2:]), axis=1)
        info.to_hdf(tmp_db, "info", "r+")
    f.close()
    return tmp_db


def make_db_for_each_file(root_directory,
                          extract_func=default_extract_func,
                          extension_filter=is_audio_file,
                          n_cores=cpu_count()):
    root_name, tree = fs_dict(root_directory, extension_filter)
    args = [(os.path.join(dir, file), extract_func)
            for dir, files in tree.items() for file in files]
    with Pool(n_cores) as p:
        tmp_dbs = p.starmap(file_to_db, args)
    return tmp_dbs


def collect_infos(tmp_dbs):
    infos = []
    for db in tmp_dbs:
        infos += [Database(db).info]
    return pd.concat(infos, ignore_index=True)


def collect_scores(tmp_dbs):
    scores = []
    offset = 0
    for db in tmp_dbs:
        scr = Database(db).score
        scr.loc[:, ("start", "stop")] = scr.loc[:, ("start", "stop")].values + offset
        scr.loc[:, "name"] = ".".join(db.split(".")[:-1])
        scores += [scr]
        offset = scr.last_stop
    return pd.DataFrame(pd.concat(scores, ignore_index=True))


def zip_prev_next(iterable):
    return zip(iterable[:-1], iterable[1:])


def ds_definitions_from_infos(infos):
    tb = infos.iloc[:, 2:].T
    paths = ["/".join(parts) for parts in infos.iloc[:, :2].values]
    # change the paths' extensions
    paths = [".".join(path.split(".")[:-1]) + ".h5" for path in paths]
    features = set(tb.index.get_level_values(0))
    ds_definitions = {}
    for f in features:
        dtype = tb.loc[(f, "dtype"), :].unique().item()
        shapes = tb.loc[(f, "shape"), :].values
        dims = shapes[0][1:]
        assert all(shp[1:] == dims for shp in
                   shapes[1:]), "all features should have the same dimensions but for the first axis"
        layout = Score.from_duration([s[0] for s in shapes])
        ds_shape = (layout.last_stop, *dims)
        layout.index = paths
        ds_definitions[f] = {"shape": ds_shape, "dtype": dtype, "layout": layout}
    return ds_definitions


def create_datasets_from_defs(target, defs, mode="w"):
    with h5py.File(target, mode) as f:
        for name, params in defs.items():
            f.create_dataset(name, shape=params["shape"], dtype=params["dtype"])
            layout = params["layout"]
            layout.reset_index(drop=False, inplace=True)
            layout = layout.rename(columns={"index": "name"})
            pd.DataFrame(layout).to_hdf(target, "layouts/" + name, "r+")
        f.close()
    return


def make_integration_args(target):
    args = []
    with h5py.File(target, "r") as f:
        for feature in f["layouts"].keys():
            df = Score(pd.read_hdf(target, "layouts/" + feature))
            args += [(target, source, feature, indices) for source, indices in
                     zip(df.name, df.slices(time_axis=0))]
    return args


def integrate(target, source, key, indices):
    with h5py.File(source, "r") as src:
        data = src[key][()]
    with h5py.File(target, "r+") as trgt:
        trgt[key][indices] = data
    return


def aggregate_dbs(target, dbs, mode="w", remove_sources=False):
    infos = collect_infos(dbs)
    score = collect_scores(dbs)
    definitions = ds_definitions_from_infos(infos)
    create_datasets_from_defs(target, definitions, mode)
    args = make_integration_args(target)
    for arg in args: integrate(*arg)
    if remove_sources:
        for src in dbs: os.remove(src)
    infos.to_hdf(target, "info", "r+")
    score.to_hdf(target, "score", "r+")


def make_root_db(db_name, root_directory, extension_filter, extract_func=default_extract_func, n_cores=cpu_count(), remove_sources=True):
    dbs = make_db_for_each_file(root_directory, extract_func, extension_filter, n_cores)
    aggregate_dbs(db_name, dbs, "w", remove_sources)


import numpy as np
from alcokit.score import Score
import unittest


class TestScore(unittest.TestCase):
    def test_constructors(self):
        sc = Score.from_stop([3, 6, 9, 12])

        assert sc.start[0] == 0
        assert sc.stop[3] == 12
        assert sc.stop[0] == 3

        ######################

        sc = Score.from_duration([3, 6, 9, 12])

        assert sc.start[0] == 0
        assert sc.stop[3] == 30
        assert sc.stop[0] == 3

        ######################

        sc = Score.from_frame_definition(20, 4, 2)
        assert sc.start[8] == 16
        assert sc.stop[8] == 20
        assert np.all(sc.duration.values == 4)
        assert sc.duration.sum() == 36

        ######################

        X = np.c_[np.random.rand(64, 100), np.ones((64, 50)), np.random.rand(64, 50)]
        sc = Score.from_recurrence_matrix(X)
        assert sc.start[0] == 0
        assert sc.stop.iloc[-1] == 200
        assert sc.duration.sum() == 200

    def test_methods(self):
        ######################

        sc = Score.from_stop([3, 6, 9, 12])

        assert sc.is_consistent()
        assert sc.is_contiguous()
        sc.stop += 2
        assert not sc.is_consistent()
        assert not sc.is_contiguous()

        sc.update_duration()
        assert sc.is_consistent()
        sc.make_contiguous()
        assert sc.is_contiguous()

        ######################

        sc = Score.from_stop([3, 6, 9, 12])
        sc.add_params(stretch=[1., 2., 1., 2.],
                      shift=[1., -1., 1., -1.])
        assert "stretch" in sc.columns
        assert "shift" in sc.columns

        ######################

    def test_join(self):
        sc = Score.from_stop([3, 6, 9, 12])
        prior_N = sc.span
        sc = sc.join_events((0, 1), (2, 3))
        assert sc.span == prior_N
        assert sc.is_contiguous()
        assert sc.is_consistent()
        assert len(sc) == 2

    def test_split(self):
        sc = Score.from_stop([3, 6, 9, 12])
        prior_span = sc.span
        prior_length = len(sc)
        sc = sc.split_events((0, 2), (1, .3333), (2, [1, 3]))

        assert len(sc) > prior_length
        assert type(sc) is Score
        assert sc.span == prior_span
        assert sc.is_contiguous()
        assert sc.is_consistent()

    def test_shift(self):
        # SPAN DESTRUCTIVE / CHAIN CONSERVING SHIFTS

        sc = Score.from_stop([3, 6, 9, 12])
        prior_span = sc.span
        sc = sc.shift_events(slice(0, None), 1)
        assert sc.span == prior_span - 1
        assert sc.is_contiguous()
        assert sc.is_consistent()

        sc = Score.from_stop([3, 6, 9, 12])
        sc = sc.shift_events(slice(0, None), -2)

        assert sc.span == prior_span - 2
        assert sc.is_contiguous()
        assert sc.is_consistent()

        # SPAN CONSERVING / CHAIN DESTRUCTIVE SHIFTS (with x_ and _x)

        sc = Score.from_stop([3, 6, 9, 12])
        sc = sc.shift_x(slice(0, None), 2, False)

        assert sc.span == prior_span
        assert not sc.is_contiguous()
        assert sc.is_consistent()

        sc = Score.from_stop([3, 6, 9, 12])
        sc = sc.x_shift(slice(0, None), -2, False)

        assert sc.span == prior_span
        assert not sc.is_contiguous()
        assert sc.is_consistent()

        # SPAN CONSERVING / CHAIN CONSERVING SHIFTS (with x_ and _x)

        sc = Score.from_stop([3, 6, 9, 12])
        sc = sc.shift_x(slice(0, None), 2, True)

        assert sc.span == prior_span
        assert sc.is_contiguous()
        assert sc.is_consistent()

        sc = Score.from_stop([3, 6, 9, 12])
        sc = sc.x_shift(slice(0, None), -2, True)

        assert sc.span == prior_span
        assert sc.is_contiguous()
        assert sc.is_consistent()

    def test_extend(self):
        sc1 = Score.from_stop([3, 6, 9, 12])
        span1 = sc1.span
        sc2 = Score.from_stop([3, 6, 9, 12])
        span2 = sc2.span

        sc = sc1.extend(sc2)

        assert sc.span == span1 + span2
        assert len(sc) == len(sc1) + len(sc2)

    def test_queries(self):
        sc = Score.from_stop([3, 6, 9, 12])
        sc.add_params(stretch=[1., 2., 1., 2.],
                      shift=[1., -1., 1., -1.])

        subset = sc.hard_q.or_(stretch=lambda s: s > 1.,
                               shift=lambda s: s > 0)
        assert (subset.index == sc.index).all()

        subset = sc.hard_q.and_(stretch=lambda s: s > 1.,
                                shift=lambda s: s > 0)
        assert len(subset) == 0


if __name__ == '__main__':
    unittest.main()



import torch
import torch.nn as nn
from alcokit.models.modules import ParamedSampler


class GatedConv(nn.Module):
    def __init__(self, dim, kernel, groups, stride=1, dilation=1, padding=0, down=True):
        super(GatedConv, self).__init__()
        mod = nn.Conv1d if down else nn.ConvTranspose1d
        self.conv_f = mod(dim, dim, kernel, stride=stride, dilation=dilation, padding=padding, groups=groups)
        self.conv_g = mod(dim, dim, kernel, stride=stride, dilation=dilation, padding=padding, groups=groups)
        self.tanh = nn.Tanh()
        self.sig = nn.Sigmoid()
        self.kernel = kernel
        self.dilation = dilation
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        f = self.tanh(self.conv_f(x))
        g = self.sig(self.conv_g(x))
        return f * g


class TimeResidualBlock(nn.ModuleList):
    def __init__(self, *modules):
        super(TimeResidualBlock, self).__init__(modules)

    def forward(self, x):
        for conv in self:
            h = conv(x[:, :, :-conv.dilation * conv.stride])
            h = torch.repeat_interleave(h, torch.tensor([conv.stride] * h.size(-1)), dim=-1)
            x = torch.cat((x[:, :, :conv.kernel * conv.dilation],
                           h + x[:, :, conv.kernel * conv.dilation:]), dim=-1)
        return x


def permute_time_freq(x):
    return x.permute(0, 2, 1).contiguous()


class ConvBlock(nn.Module):
    def __init__(self, dim, n_layers=1, kernel=2, groups=1, stride=1, dilated=False, down=True, forward="sequential"):
        super(ConvBlock, self).__init__()
        self.forward_method = forward
        if self.forward_method == "sequential":
            self.block = nn.Sequential(
                *[GatedConv(dim, kernel, groups, stride=stride,
                            dilation=2 ** i if dilated else 1, down=down) for i in range(n_layers)]
            )
        elif self.forward_method == "residual":
            self.block = TimeResidualBlock(
                *[GatedConv(dim, kernel, groups, stride=stride,
                            dilation=2 ** i if dilated else 1, down=down) for i in range(n_layers)]
            )
        else:
            raise ValueError("`forward` argument not recognized. Must be one of 'sequential', 'residual'")

    def forward(self, x):
        """
        input shape : (batch_size x T x model_dim)
        output shape : (batch_size x T' x model_dim)
        """
        x = permute_time_freq(x)
        x = self.block(x)
        x = permute_time_freq(x)
        return x

    def receptive_field(self):
        modules = self.block if isinstance(self.block, TimeResidualBlock) else list(self.block.children())
        rf = modules[0].kernel
        for cv in modules[1:]:
            if cv.stride == cv.kernel:
                rf *= (cv.kernel - 1) * cv.stride
            else:
                rf += (cv.kernel - 1) * cv.dilation
        return rf


class ConvolutionNetwork(nn.Sequential):
    def __init__(self, dim, symmetrical=False, core=ParamedSampler, **conv_blocks_kwargs):
        if symmetrical:
            if "down" in conv_blocks_kwargs:
                conv_blocks_kwargs.pop("down")
            super(ConvolutionNetwork, self).__init__(
                ConvBlock(dim, **conv_blocks_kwargs, down=True),
                ParamedSampler(dim, dim, return_params=False),
                ConvBlock(dim, **conv_blocks_kwargs, down=False)
            )
        else:
            super(ConvolutionNetwork, self).__init__(
                ConvBlock(dim, **conv_blocks_kwargs)
            )

    def receptive_field(self):
        return next(self.children()).receptive_field()

    def future_length(self):
        n_children = len(list(self.children()))
        if n_children > 1:
            return self.receptive_field()
        return 1


import numpy as np
import scipy.stats as stats
import librosa
from alcokit.util import normalize
from alcokit.transform.spectrum import frame, running_agg
from librosa.segment import recurrence_matrix
from librosa.util import localmax
from scipy.ndimage import convolve


def from_recurrence_matrix(X,
                           L=6,
                           k=None,
                           sym=True,
                           bandwidth=1.,
                           thresh=0.2,
                           min_dur=4):
    def checker(N):
        block = np.zeros((N * 2 + 1, N * 2 + 1), dtype=np.int32)
        for k in range(-N, N + 1):
            for l in range(-N, N + 1):
                block[k + N, l + N] = np.sign(k) * np.sign(l)
        return block / abs(block).sum()

    R = recurrence_matrix(
        X, metric="cosine", mode="affinity",
        k=k, sym=sym, bandwidth=bandwidth, self=True)
    # intensify checker-board-like entries
    R_hat = convolve(R, checker(L),
                     mode="constant")  # Todo : check if mode="reflect" would avoid peaking at the end
    # extract them along the main diagonal
    dg = np.diag(R_hat, 0)
    mx = localmax(dg * (dg > thresh)).nonzero()[0]
    # filter out maxes less than min_dur frames away of the previous max
    mx = mx * (np.diff(mx, append=R.shape[0]) >= min_dur)
    mx = mx[mx > 0]
    stops = np.r_[mx, R.shape[0]]
    return stops


def as_novelty_f(graph, width=3, order=1, axis=0):
    if width < 3:
        grad = np.diff(normalize(graph), n=width, prepend=np.zeros(width))
    else:
        grad = librosa.feature.delta(normalize(graph), width=width, order=order, axis=axis)
    grad[grad < 0] = 0
    return normalize(grad)


def reduce_sequence(data, n=1, t_axis=-1, agg=lambda x, axis: x, J=None):
    agged = running_agg(data, agg=agg, window=n, hop_length=1, mode="edge",
                        p_axis=t_axis, f_axis=t_axis, a_axis=1)
    T = data.shape[t_axis]
    costs = np.zeros(T)
    if t_axis == 0:
        for t in range(n, T):
            costs[t] = J(data, agged, t, n)
    else:
        for t in range(n, T):
            costs[t] = J(data.T, agged.T, t, n)
    return costs


def KL(P, Q, t, n):
    if P[t].sum() < 5e-5 or Q[t-n].sum() < 5.e-5:
        return 0.  # ignore frames where the sum of energy is below 5.e-5
    else:
        return stats.entropy(P[t], Q[t-n])


def KL_update(P, Q, t, n):
    prior = np.mean(Q[t-n], axis=0)
    posterior = np.mean(np.vstack((Q[t-n], P[t])), axis=0)
    if prior.sum() > 5e-5 or posterior.sum() < 5.e-5:
        return 0.  # ignore frames where the sum of energy is below 5.e-5
    else:
        return stats.entropy(posterior, prior)


def self_perplexity(data, n=1, t_axis=-1, agg=np.mean, width=3, order=1):
    perplexity = reduce_sequence(data, n=n, t_axis=t_axis, agg=agg, J=KL)
    return as_novelty_f(perplexity, width=width, order=order, axis=0)


def self_perplexity_update(data, n=1, t_axis=-1, agg=lambda x, axis: x, width=3, order=1):
    perplexity = reduce_sequence(data, n=n, t_axis=t_axis, agg=agg, J=KL_update)
    return as_novelty_f(perplexity, width=width, order=order, axis=0)


# Peak Picking :

def max_in_range(y, rg=4, hop=1):
    framed = frame(y, m_frames=rg, hop_length=hop, p_axis=0, f_axis=0)
    idx = (framed.argmax(axis=1) + (np.arange(framed.shape[0]) + 1 - (rg//2)))[framed.max(axis=1) > 0]
    items, counts = np.unique(idx, return_counts=True)
    return items[counts >= rg]


def mad_outliers(y,thresh=3.5):
    """
    find outliers with median absolute deviation
    taken from https://stackoverflow.com/questions/22354094/pythonic-way-of-detecting-outliers-in-one-dimensional-observation-data
    """
    m = max(np.median(y), 1e-6)
    abs_dev = np.abs(y - m)
    left_mad = max(np.median(abs_dev[y <= m]), 1e-6)
    right_mad = max(np.median(abs_dev[y >= m]), 1e-6)
    y_mad = m * np.ones(len(y))
    y_mad[y > m] = right_mad
    modified_z_score = 0.6745 * abs_dev / y_mad
    modified_z_score[y == m] = 0
    modified_z_score[np.isnan(modified_z_score) | np.isinf(modified_z_score)] = 0
    return modified_z_score > thresh


# Freq Bands Splitter :
def split_bands(data, *freqs):
    F, T = data.shape
    bins_f = librosa.fft_frequencies(n_fft=(F-1) * 2, sr=22500)
    k = 0
    bands = np.zeros(bins_f.shape, dtype=np.int32)
    for i, f in enumerate(bins_f):
        if f >= freqs[k]:
            k += 1
            bands[i] = k
        else:
            bands[i] = k
    bands = np.flatnonzero(np.diff(bands))
    return np.split(data, bands, axis=0)


from alcokit import HOP_LENGTH, SR
from alcokit.transform.spectrum import griffinlim
from alcokit.util import signal
from librosa import stft, phase_vocoder
import numpy as np
from pyrubberband.pyrb import time_stretch as rb_stretch
from scipy.interpolate import RectBivariateSpline as RBS


def trim(S, left=0, right=0):
    return S[:, left:-right] if right else S[left:]


def pad(S, left=0, right=0, **kwargs):
    return np.pad(S, ((0, 0), (left, right)), **kwargs)


def split(S, indices):
    return np.split(S, indices, axis=1)


def concat(*spectra):
    return np.concatenate(spectra, axis=1)


def reverse(S):
    return np.fliplr(S)


def add_start_end_tokens(S, start, end):
    pass


def _stretch_hop(S, ratio):
    y = signal(S)
    hop = int(HOP_LENGTH * ratio)
    return stft(y, hop_length=hop)


def _stretch_rbs(S, ratio):
    if S.shape[1] <= 1 or S.shape[0] <= 1:
        return S
    time_indices = np.linspace(0, S.shape[1]-1, int(np.rint(S.shape[1]/ratio)))
    if S.dtype in (np.complex64, np.complex128):
        mag, phase = abs(S), np.imag(S)
    else:
        mag, phase = S, None
    spline = RBS(np.arange(S.shape[0]), np.arange(S.shape[1]), mag)
    interp = spline.ev(np.arange(S.shape[0])[:, None], time_indices)
    if S.dtype in (np.complex64, np.complex128):
        return griffinlim(interp, n_iter=32)
    else:
        return interp


def _stretch_vocoder(S, ratio):
    return phase_vocoder(S, ratio, hop_length=HOP_LENGTH)


def _stretch_rubberband(S, ratio):
    y = signal(S)
    return stft(rb_stretch(y, SR, ratio), hop_length=HOP_LENGTH)


def stretch(S, ratio, method="rbs"):
    funcs = {"rbs": _stretch_rbs, "hop": _stretch_hop, "rubber": _stretch_rubberband, "voc": _stretch_vocoder}
    if method not in funcs:
        raise ValueError("'method' value '{}' not understood."
                         "'method' arg should be one of ['hop', 'rbs', 'voc', 'rubber']")
    func = funcs[method]
    return func(S, ratio)


def align(X, Y, method="rbs"):
    N, M = X.shape[1], Y.shape[1]
    k = (N + M) / 2
    x, y = stretch(X, N/k, method), stretch(Y, M/k, method)
    assert x.shape[1] == y.shape[1]
    return x, y


def stretch_map(S, map, method="rbs"):
    pass




import pandas as pd
import numpy as np
import torch
from alcokit.extract.segmentation import from_recurrence_matrix


def ssts(item, axis=0):
    """
    Start-Stop To Slices
    @param item: a pd.DataFrame with a "start" and a "stop" column (typically the metadata of a db)
    @param axis:
    @return: 1d array of slices for retrieving each element in `item`
    """
    arr = np.atleast_2d(item[['start', 'stop']].values)
    slices = map(lambda a: (*[slice(0, None)] * axis, slice(a[0], a[1]), ), arr)
    return slices


class Score(pd.DataFrame):

    # OVERRIDES

    @property
    def _constructor(self):
        return Score

    @staticmethod
    def _validate(obj):
        if 'start' not in obj.columns or "stop" not in obj.columns:
            raise ValueError("Must have 'start' and 'stop' columns.")

    # PROPERTIES

    @property
    def first_start(self):
        return self.start.min()

    @property
    def last_stop(self):
        return self.stop.max()

    @property
    def span(self):
        return self.last_stop - self.first_start

    @property
    def cumdur(self):
        return np.cumsum(self["duration"].values)

    @property
    def flat(self):
        return np.array([i for ev in self.events for i in range(ev.start, ev.stop)])

    def slices(self, time_axis=0):
        """
        This is the back-end core of a score. This method efficiently returns the indexing objects
        necessary to communicate with n-dimensional data.

        @return: an array of slices where slice_i corresponds to the row/Event_i in the DataFrame
        """
        return ssts(self, time_axis)

    @property
    def durations_(self):
        return self.stop.values - self.start.values

    @property
    def ovl_defs(self):
        last = self.stop.values[-1]
        ovl = np.array([(a, b - a)
                        for a, b in zip(self.start.values[1:], self.stop.values[:-1])]
                       + [(last, 0)])
        return ovl

    @property
    def events(self):
        return self.itertuples(name="Event", index=True)

    # UPDATING METHOD

    def make_contiguous(self):
        self.reset_index(drop=True, inplace=True)
        cumdur = self.cumdur
        self.loc[:, "start"] = np.r_[0, cumdur[:-1]] if cumdur[0] != 0 else cumdur[:-1]
        self.loc[:, "stop"] = cumdur
        return self

    def update_duration(self):
        self.reset_index(drop=True, inplace=True)
        self.loc[:, "duration"] = self.stop.values - self.start.values
        return self

    def clean_up(self):
        """remove Events with 0 duration"""
        self.drop(index=(self.duration.values == 0).nonzero()[0], inplace=True)
        self.reset_index(drop=True, inplace=True)
        return self

    # Sanity Checks

    def is_consistent(self):
        return (self["duration"].values == (self.stop.values - self.start.values)).all()

    def is_contiguous(self):
        return (self.start.values[1:] == self.stop.values[:-1]).all()

    # MUTATING METHODS

    def add_params(self, **kwargs):
        for name, val in kwargs.items():
            self.loc[:, name] = val
        return self

    def add_duration(self):
        return self.update_duration()

    def add_stretch(self, default=1.):
        return self.add_params(stretch=default)

    def add_stretch_from_target_duration(self, targets):
        targets = np.asarray(targets)
        return self.add_stretch(default=self.durations_ / targets)

    def add_shift(self, default=0.):
        return self.add_params(shift=default)

    def add_target_duration(self):
        if "duration" not in self.columns or "stretch" not in self.columns:
            raise ValueError("Must contain the columns 'duration' and 'stretch' to compute 'target_duration'")
        self.loc[:, "target_duration"] = self.durations_ / self.stretch.values
        return self

    def cut_at(self, *frame_i):
        """
        adds events whose stop are at each frame_i and starts at the first stop < frame_i
        @param frame_i:
        @return:
        """
        pass

    def uncut_at(self, *event_i):
        """
        remove the events with indices `events_i` by merging with them with the next one
        @param event_i:
        @return:
        """
        pass

    def join_events(self, *event_i):
        for i in event_i:
            first = i.start if type(i) is slice else min(i)
            others = list(range(i.start + 1, i.stop)) if type(i) is slice else list(set(i) - {first})
            self.loc[first, "stop"] = self.loc[others, "stop"].max()
            if "duration" in self.columns:
                self.loc[first, "duration"] = self.loc[first, "stop"] - self.loc[first, "start"]
            self.drop(index=others, inplace=True)
        self.reset_index(drop=True, inplace=True)
        return self

    def split_events(self, *splits_def):
        """
        @param sp_def: some iterable of iterables where :
                - the first element is the index of the event to split,
                - the rest of the iterable can be:
                    - an int: number of splits of equal length to perform (any reminder will be added to the last split)
                    - a float: floor(1 / x) splits of equal length will be performed (the reminder will be added to the last split)
                    - a list/iterable of stops: will split the event at those specific locations
        @return:
        """

        def _as_stops(sp_def, duration):
            if getattr(sp_def, "__iter__", False):
                return np.asarray(sorted(list(set(sp_def) | {duration})))
            elif type(sp_def) is int:
                eq_n = [duration // sp_def] * sp_def
                eq_n[-1] += duration % sp_def
                return np.cumsum(eq_n)
            elif type(sp_def) is float and sp_def < 1.:
                n = int(np.floor(1 / sp_def))
                eq_n = [duration // n] * n
                eq_n[-1] += duration % n
                return np.cumsum(eq_n)
            else:
                raise ValueError("split_definition '{}' not recognized".format(sp_def))

        splits_def = sorted(list(splits_def), key=lambda s: s[0])
        i_offset = 0
        for s_def in splits_def:
            index = i_offset + s_def[0]
            # build the new score to be inserted
            to_split = self.loc[index, :]
            dur = to_split.stop - to_split.start
            stops = _as_stops(s_def[1], dur)
            new = Score.from_stop(stops + to_split.start)
            new["stop"] = stops + to_split["start"]
            new["start"] = np.r_[0, stops[:-1]] + to_split["start"]
            if "duration" in self.columns:
                new = new.add_duration()
            # concat everything and reset indices
            prior = pd.concat((self.iloc[:index], new))
            self = pd.concat((prior, self.iloc[index + 1:]),
                             axis=0,
                             ignore_index=True)
            i_offset += len(stops) - 1
        return self

    def shift_events(self, event_i, n, adjust_others=True):
        self.shift_x(event_i, n, False)
        self.x_shift(event_i, n, False)
        if adjust_others:
            diffs = np.r_[self.start.values[1:] - self.stop.values[:-1], 0]
            if n > 0:
                self.loc[diffs > 0, "stop"] = self.loc[diffs > 0, "stop"].values + diffs[diffs > 0]
                diffs = np.roll(diffs, 1, axis=0)
                self.loc[diffs < 0, "start"] = self.loc[diffs < 0, "start"].values - diffs[diffs < 0]
            else:
                self.loc[diffs < 0, "stop"] = self.loc[diffs < 0, "stop"].values + diffs[diffs < 0]
                diffs = np.roll(diffs, 1, axis=0)
                self.loc[diffs > 0, "start"] = self.loc[diffs > 0, "start"].values - diffs[diffs > 0]
        if "duration" in self.columns:
            self.update_duration()
            self.clean_up()
        return self

    def x_shift(self, event_i, n, adjust_x=True):
        event_i = np.asarray(event_i) if type(event_i) is not slice \
            else np.arange(len(self))[event_i]
        starts = self.loc[event_i, "start"].values
        self.loc[event_i, "start"] = np.maximum(n + starts, self.first_start)
        if adjust_x:
            self.shift_x(np.maximum(event_i - 1, 0), n, x_adjust=False)
        if "duration" in self.columns:
            self.update_duration()
            self.clean_up()
        return self

    def shift_x(self, event_i, n, x_adjust=True):
        event_i = np.asarray(event_i) if type(event_i) is not slice \
            else np.arange(len(self))[event_i]
        stops = self.loc[event_i, "stop"].values
        self.loc[event_i, "stop"] = np.minimum(n + stops, self.last_stop)
        if x_adjust:
            self.x_shift(np.minimum(event_i + 1, len(self) - 1), n, adjust_x=False)
        if "duration" in self.columns:
            self.update_duration()
            self.clean_up()
        return self

    def overlap(self, event_i, n):
        """
        combines shift_x(event_i, n) and x_shift(event_i, -n)
        @param event_i:
        @param n:
        @return:
        """
        self.shift_x(event_i, n, x_adjust=False)
        self.x_shift(event_i, -n, adjust_x=False)
        return self

    def overlap_x(self, event_i, n):
        return self.shift_x(event_i, n, x_adjust=False)

    def x_overlap(self, event_i, n):
        return self.x_shift(event_i, -n, adjust_x=False)

    def extend(self, other):
        Score._validate(other)
        other.loc[slice(0, None), ("start", "stop")] = self.last_stop + other.loc[
            slice(0, None), ("start", "stop")].values
        return pd.concat([self, other], ignore_index=True, axis=0)

    # CONSTRUCTORS

    @staticmethod
    def from_start_stop(starts, stops, durations):
        return Score((starts, stops, durations), index=["start", "stop", "duration"]).T

    @staticmethod
    def from_stop(stop):
        """
        integers in `stops` correspond to the prev[stop] and next[start] values.
        `stops` must contain the last index ! and it can begin with 0, but doesn't have to...
        """
        stop = np.asarray(stop)
        if stop[0] == 0:
            starts = np.r_[0, stop[:-1]]
            stop = stop[1:]
        else:
            starts = np.r_[0, stop[:-1]]
        durations = stop - starts
        return Score.from_start_stop(starts, stop, durations)

    @staticmethod
    def from_duration(duration):
        stops = np.cumsum(duration)
        starts = np.r_[0, stops[:-1]]
        return Score.from_start_stop(starts, stops, duration)

    @staticmethod
    def from_data(sequence):
        if issubclass(type(sequence[0]), np.ndarray):
            axis = 1
        elif issubclass(type(sequence[0]), torch.Tensor):
            axis = 0
        else:
            raise TypeError(
                "sequence's elements' type not recognized. Must be a subclass of np.ndarray or torch.Tensor")
        duration = np.array([x.shape[axis] for x in sequence])
        return Score.from_duration(duration)

    @staticmethod
    def from_frame_definition(total_duration, frame_length, stride=1, butlasts=0):
        starts = np.arange(total_duration - frame_length - butlasts + 1, step=stride)
        durations = frame_length + np.zeros_like(starts, dtype=np.int)
        stops = starts + durations
        return Score.from_start_stop(starts, stops, durations)

    @staticmethod
    def from_recurrence_matrix(X, **kwargs):
        stops = from_recurrence_matrix(X, **kwargs)
        return Score.from_stop(stops)


@pd.api.extensions.register_dataframe_accessor("soft_q")
class SoftQueryAccessor:
    def __init__(self, pandas_obj):
        self._df = pandas_obj

    def or_(self, **kwargs):
        series = False
        for col_name, func in kwargs.items():
            series = series | func(self._df[col_name])
        return series

    def and_(self, **kwargs):
        series = True
        for col_name, func in kwargs.items():
            series = series & func(self._df[col_name])
        return series


@pd.api.extensions.register_dataframe_accessor("hard_q")
class HardQueryAccessor:
    def __init__(self, pandas_obj):
        self._df = pandas_obj

    def or_(self, **kwargs):
        series = self._df.soft_q.or_(**kwargs)
        return self._df[series].reset_index(drop=True)

    def and_(self, **kwargs):
        series = self._df.soft_q.and_(**kwargs)
        return self._df[series].reset_index(drop=True)



from alcokit.models.data import *
from alcokit.hdf.factory import file_to_db
from alcokit.hdf.api import Database
from alcokit.score import Score
import unittest
import os

# Test data

X_np = np.random.rand(333, 111)
Y_np = np.random.rand(333, 3)
X_torch = torch.from_numpy(X_np)
Y_torch = torch.from_numpy(Y_np)

score = Score.from_duration([111, 111, 111])
subscore = score.iloc[[0, 2]].copy()


# Test DB

def extract_func(path):
    return {"X": ({}, X_np), "Y": ({}, Y_np), "score": ({}, score)}


db_path = file_to_db("testdb", extract_func, "w")
db = Database(db_path)
X_h5 = db.X
Y_h5 = db.Y


class TestData(unittest.TestCase):

    def test_flat_sampler(self):
        smp = FlatSampler(score)

        # contiguous score
        assert len(smp) == 333
        assert len([i for i in smp]) == 333

        # discontinuous score
        smp = FlatSampler(subscore)
        assert len(smp) == 222
        assert len([i for i in smp]) == 222

    def test_event_sampler(self):
        smp = EventSampler(score)
        idx = [i for i in smp]

        # contiguous score
        assert len(smp) == 3
        assert len(idx) == 3
        assert all("Event" in str(type(i)) for i in smp)

        # discontinuous score
        smp = EventSampler(subscore)
        idx = [i for i in smp]

        assert len(smp) == 2
        assert len(idx) == 2
        assert all("Event" in str(type(i)) for i in smp)

    def test_frame_sampler(self):
        smp = FrameSampler(N=333, k=5, stride=1, shifts=(0, 4))
        idx = [i for i in smp]

        assert len(smp) == 333-5-4+1
        assert len(idx) == 333-5-4+1
        assert all(type(i) is slice for i in smp)

    def test_numpy_set(self):
        pass
        # Single Feature

        # Multi Feature

    def test_torch_set(self):
        pass

    def test_opened_ds(self):
        pass

    def test_prepare(self):
        pass

    def test_load(self):
        # Single feature np
        dl = load(X_np, score, "flat",
                  batch_size=32,
                  pre_cat=False, collate_fn=zip_stack)

        N = sum([b[0].size(0) for b in dl])
        assert N == 333, N

        # Single feature HDF
        dl = load(X_h5, score, "flat",
                  batch_size=32,
                  pre_cat=False, collate_fn=zip_stack)

        N = sum([b[0].size(0) for b in dl])
        assert N == 333, N

        # Single feature HDF PRECAT SUBSCORE
        dl = load(X_h5, subscore, "flat",
                  batch_size=32,
                  pre_cat=True, collate_fn=zip_stack)

        N = sum([b[0].size(0) for b in dl])
        assert N == 222, N

        # MULTI feature HDF
        dl = load((X_h5, Y_h5), score, "flat",
                  batch_size=32,
                  pre_cat=False, collate_fn=zip_stack)
        xx = [b for b in dl][0]
        N = sum([b[0].size(0) for b in dl])
        assert N == 333, N

        # Single feature HDF
        dl = load(X_h5, score, "event",
                  batch_size=32,
                  pre_cat=False, collate_fn=zip_stack)

        N = sum([b[0].size(0) for b in dl])
        assert N == 3, N

        # MULTI feature HDF
        dl = load((X_h5, Y_h5), score,
                  "frame", k=11, stride=1, shifts=(0, 3),
                  batch_size=32,
                  pre_cat=False, collate_fn=zip_stack)
        xx = [b for b in dl][0]
        N = sum([b[0].size(0) for b in dl])
        assert N == 333-11-3+1, N


if __name__ == '__main__':
    unittest.main()

    # clean up
    os.remove(db_path)


from alcokit import SR, HOP_LENGTH
from alcokit.util import signal, f2s
import numpy as np
from librosa import resample, phase_vocoder, util, stft
from pyrubberband.pyrb import pitch_shift as rb_shift


def _shift_translate(S, k_bins):
    n = S.shape[0]
    k = k_bins
    rng = np.arange(n)
    i_rng = rng if k == 0 else (rng[:-k] if k > 0 else rng[-k:])
    j_rng = rng if k == 0 else (rng[k:] if k > 0 else rng[:k])
    D = np.zeros((n, n))
    D[i_rng, j_rng] = 1
    return D.T @ S


def _shift_scale(S, scale):
    rate = 1 / scale
    n = S.shape[0]
    i_range = np.arange(n)
    j = i_range * rate
    D = np.zeros((n, n))

    for i, alpha in zip(i_range, j):
        floor = int(np.floor(alpha))
        ceil = floor + 1
        if floor >= n:
            break
        col = [floor, ceil] if ceil < n else floor
        val = [ceil - alpha, alpha - floor] if ceil < n else ceil - alpha
        D[i, col] = val
    return D.T @ S


def steps2rate(n_steps, bins_per_octave=12):
    if bins_per_octave < 1 or not np.issubdtype(type(bins_per_octave), np.integer):
        raise ValueError('bins_per_octave must be a positive integer.')
    return 2.0 ** (-float(n_steps) / bins_per_octave)


def rate2steps(rate):
    return np.log2(rate) * -12


def _shift_vocoder(S, rate):
    y = signal(phase_vocoder(S, rate, hop_length=HOP_LENGTH))
    y_shift = resample(y, float(SR) / rate, SR,
                       res_type="kaiser_best")

    # Crop to the same dimension as the input
    return stft(util.fix_length(y_shift, f2s(S.shape[1], HOP_LENGTH)))


def _shift_rubber(S, intv):  # TODO : DOES THIS SUPPORT QUARTER-TONES ??
    y = signal(S)
    return stft(rb_shift(y, SR, rate2steps(intv)), hop_length=HOP_LENGTH)


def shift(S, intv, method="voc"):
    funcs = {"trans": _shift_translate, "scale": _shift_scale, "rubber": _shift_rubber, "voc": _shift_vocoder}
    if method not in funcs:
        raise ValueError("'method' value '{}' not understood."
                         "'method' arg should be one of ['trans', 'scale', 'voc', 'rubber']")
    func = funcs[method]
    return func(S, intv)


def rotate(S, k, clip=True):
    pass


def flip(S):
    return np.flipud(S)


def repitch(S):
    pass


def rephase(S, eps):
    pass



import numpy as np
from scipy.stats import rv_histogram
from scipy.ndimage import generic_filter
from alcokit.transform.spectrum import frame


# Low level functions
def categorical_split(X, n_bins=200, min_size=100, ignore_zeros=True):
    """
    fit an empirical distribution D to X and return the points in the domain of X
    where the sign of the derivative of the pdf of D is 0
    and where the value of this derivative is maximum.
    This corresponds to partionning X into n sub-domains where the pdf of the sub-domain_i
    is best approximated by a uniform distribution.
    """
    x_min, x_max = X.min(), X.max()
    if X.size < min_size or (x_max - x_min) == 0:
        return np.r_[x_min, x_min, x_max]
    hist = np.histogram(X, bins=n_bins)
    dist = rv_histogram(hist)
    xs = np.linspace(x_min, x_max, n_bins)
    uniform_cdf = xs * (1 / (xs[-1] - xs[0]))
    uniform_cdf -= uniform_cdf.min()
    dev = uniform_cdf - dist.cdf(xs)
    dev_sign = np.sign(dev)
    dev_abs = abs(dev)
    # X is already "almost" uniform
    if dev_abs.max() <= 1e-1:  # OR np.std(dev_abs) ?
        sp = (x_min, x_min, x_max)
    # X hasn't any zero crossings
    elif ignore_zeros or (np.all(dev_sign[1:-1] >= 0) or np.all(dev_sign[1:-1] <= 0)):
        sp = (x_min, xs[dev_abs.argmax()], x_max)
    # return Zero crossings and their respective maxes
    else:
        diffs = abs(np.diff(dev_sign[1:-1], prepend=0))
        zeros = np.where(diffs == 2)[0] + 1
        maxes = [x.argmax() + i for i, x in zip(np.r_[0, zeros], np.split(dev_abs, zeros))]
        sp = (x_min, *tuple(np.sort(xs[np.r_[zeros, maxes]])), x_max)
    return np.r_[sp]


def csplit_right(X, splits, **kwargs):
    new = categorical_split(X[X >= splits[-2]], **kwargs)
    return np.r_[splits[:-1], new[1:]]


def csplit_left(X, splits, **kwargs):
    new = categorical_split(X[X <= splits[1]], **kwargs)
    return np.r_[new[:-1], splits[1:]]


def k_csplits(X, k, side="right", **kwargs):
    splits = categorical_split(X, **kwargs)
    if side == "right":
        for n in range(max([0, k - 2])):
            splits = csplit_right(X, splits, **kwargs)
    else:
        for n in range(max([0, k - 2])):
            splits = csplit_left(X, splits, **kwargs)
    return splits


def csplit_along_axis(X, k, side="right", axis=None, **kwargs):
    if axis is None:
        return k_csplits(X.flat[:], k, side, **kwargs)
    return np.apply_along_axis(k_csplits, int(not axis), X, k, side, **kwargs)


def tag(X, csplits):
    tags = np.zeros_like(X, dtype=np.int32) - 1
    if len(csplits.shape) == 1:
        return np.sum(np.stack(tuple(X >= s for s in csplits[:-1])), axis=0, out=tags) - 1
    else:
        matching_dim = (np.r_[X.shape] == np.r_[csplits.shape]).nonzero()[0][0]
        func = lambda s: X >= (s if matching_dim == 1 else s[:, None])
        return np.sum(np.stack(tuple(func(s) for s in (csplits if matching_dim == 1 else csplits.T))),
                      axis=0, out=tags) - 1


class SupportTagger(object):
    def __init__(self, data, k=2, ignore_zeros=True, side="right"):
        self.ignore_zeros = ignore_zeros
        self.side = side
        self.splits = k_csplits(data.flat[:], k, side, ignore_zeros=ignore_zeros)

    def tag(self, X):
        return tag(X, self.splits)


def smooth_tags(tags, axis=None, kernel=None, window=2, agg=np.mean):
    if axis is not None and np.any(kernel is not None):
        raise ValueError("either axis or kernel has to be None")
    if axis is not None:
        framed = frame(tags, window, 1, "reflect", p_axis=axis, f_axis=axis)
        return agg(framed, axis=-1 if axis == 0 else 0)
    if np.any(kernel):
        return generic_filter(tags, agg, footprint=kernel)
    else:
        raise ValueError("both axis and kernel are None. Can not compute anything.")


def discrete_affinity(x, ref, glob_min, glob_max):
    """
    return the absolute affinity from x to ref
    as a proportion of (ref - glob_min) if x < ref
    or as a proportion of (glob_max - x) if x > ref.
    useful for mixing tagged elements in an array.
    """
    if x < glob_min or x > glob_max:
        return 0
    if (ref - x) > 0:
        return 1 - (ref - x) / (ref - glob_min)
    if (ref - x) < 0:
        return 1 - (x - ref) / (glob_max - ref)
    # ref == x
    return 1


def mix(X, tags, affinity_functions=(), normalize=True):
    K = len(affinity_functions)
    if K == 0:
        raise ValueError("the argument affinity_functions is empty")
    if X.shape != tags.shape:
        raise ValueError("X and tags must have the same shape")
    filtr = np.zeros_like(X, dtype=X.dtype)
    for func in affinity_functions:
        filtr += func(tags)
    return X * (filtr / (K if normalize else 1))



import librosa
import numpy as np
import os
from librosa.display import specshow
import matplotlib.pyplot as plt
import IPython.display as ipd
from alcokit import HOP_LENGTH, SR, N_FFT
import pickle


def save_pickle(obj, path):
    with open(path, "wb") as f:
        f.write(pickle.dumps(obj))
    return None


def load_pickle(path):
    with open(path, "rb") as f:
        obj = pickle.loads(f.read())
    return obj


# OS
def is_audio_file(file):
    return file.split(".")[-1] in ("wav", "aif", "aiff", "mp3", "m4a", "mp4") and "._" not in file


def flat_dir(directory, ext_filter=is_audio_file):
    files = []
    for root, dirname, filenames in os.walk(directory):
        for f in filenames:
            if ext_filter(f):
                files += [os.path.join(root, f)]
    return sorted(files)


def fs_dict(root, extension_filter=is_audio_file):
    root_name = os.path.split(root.strip("/"))[-1]
    items = [(d, list(filter(extension_filter, f))) for d, _, f in os.walk(root)]
    if not items:
        raise ValueError("no audio files found on path %s" % root)
    return root_name, dict(item for item in items if len(item[1]) > 0)


# Conversion

normalize = librosa.util.normalize
a2db = lambda S: librosa.amplitude_to_db(abs(S), ref=S.max())
s2f = librosa.samples_to_frames
s2t = librosa.samples_to_time
f2s = librosa.frames_to_samples
f2t = librosa.frames_to_time
t2f = librosa.time_to_frames
t2s = librosa.time_to_samples
hz2m = librosa.hz_to_midi
m2hz = librosa.midi_to_hz


def m2b(m, sr=SR, n_fft=N_FFT):
    step = (sr / 2) / (n_fft // 2)
    return m2hz(m) / step


def b2m(b, sr=SR, n_fft=N_FFT):
    step = (sr / 2) / (n_fft // 2)
    return hz2m(b * step)


def delta_b(b, delta_m=1, sr=SR, n_fft=N_FFT):
    """
    returns the size in bins of the interval delta_m (in midi) at bin `b`
    """
    params = dict(sr=sr, n_fft=n_fft)
    return b - m2b(b2m(b, **params) - delta_m, **params)


def unit_scale(x):
    return (x - x.min()) / (x.max() - x.min())


# Debugging utils

def db(S):
    if S.dtype == np.complex64:
        S_hat = a2db(S.abs) + 40
    elif S.min() >= 0 and S.dtype in (np.float, np.float32, np.float64, np.float_):
        S_hat = a2db(S) + 40
    else:
        S_hat = a2db(S)
    return S_hat


def signal(S, hop_length=HOP_LENGTH):
    if S.dtype in (np.complex64, np.complex128):
        return librosa.istft(S, hop_length=hop_length)
    else:
        return librosa.griffinlim(S, hop_length=hop_length, n_iter=32)


def audio(S, hop_length=HOP_LENGTH, sr=SR):
    if len(S.shape) > 1:
        y = signal(S, hop_length)
        if y.size > 0:
            return ipd.display(ipd.Audio(y, rate=sr))
        else:
            return ipd.display(ipd.Audio(np.zeros(hop_length*2), rate=sr))
    else:
        return ipd.display(ipd.Audio(S, rate=sr))


def playlist(iterable):
    for seg in iterable:
        audio(seg)
    return


def playthrough(iterable, axis=1):
    rv = np.concatenate(iterable, axis=axis)
    return audio(rv)


def show(S, figsize=(), to_db=True, y_axis="linear", x_axis='frames', title=""):
    S_hat = db(S) if to_db else S
    if figsize:
        plt.figure(figsize=figsize)
    ax = specshow(S_hat, x_axis=x_axis, y_axis=y_axis, sr=SR)
    plt.colorbar()
    plt.tight_layout()
    plt.title(title)
    return ax



import numpy as np
from sklearn.metrics import pairwise_distances
from multiprocessing import cpu_count, Pool, get_context
from scipy.sparse import issparse


def _mp_block_reduce(arr, splits_j, reduce_func, subst_zeros=None):
    X = arr.A if issparse(arr) else arr
    X[X == 0] = 1
    X = np.split(X, splits_j, axis=1)
    return np.array([reduce_func(x) for x in X])


def reduce_2d(arr_2d, splits_i, splits_j, reduce_func=np.mean, subst_zeros=None, n_cores=cpu_count()):
    Ax = [arr_2d[xi:xj] for xi, xj in zip(np.r_[0, splits_i[:-1]], splits_i)]
    args = [(ax, splits_j[:-1], reduce_func, subst_zeros) for ax in Ax]
    with Pool(n_cores) as p:
        rows = p.starmap(_mp_block_reduce, args)
    return np.stack(rows)


def mp_foreach(func, args, n_cores=cpu_count()):
    with get_context("spawn").Pool(n_cores) as p:
        res = p.starmap(func, args)
    return res


def sim_graph(x_tuple, y_tuple, metric, reduce_func=None, n_jobs=None):
    dbx, dby = x_tuple[0], y_tuple[0]
    item_x, item_y = x_tuple[1], y_tuple[1]
    slice_x, slice_y = item_x.slices(0), item_y.slices(0)
    data_x, data_y = np.concatenate(dbx[item_x], axis=int(dbx.T)), np.concatenate(dby[item_y], axis=int(dby.T))
    data_x = data_x.T if dbx.T else data_x
    data_y = data_y.T if dby.T else data_y
    G = pairwise_distances(data_x, data_y, metric=metric, n_jobs=n_jobs)
    if reduce_func is not None:
        G = np.stack([reduce_func(block, axis=0)
                      for block in np.split(G, slice_x)[:-1]])
        G = np.hstack([reduce_func(block, axis=1)[:, None]
                       for block in np.split(G, slice_y, axis=1)[:-1]])
    return G


def sort_graph(G, mode, param):
    locs = None
    if mode == "best":
        idx = np.argsort(G, axis=1)[:, :param]
        locs = [neighbors for neighbors in idx]
    elif mode == "radius":
        locs = []
        for row in G:
            idx = np.argsort(row)
            less_than = idx[row[idx] <= param]
            locs += [less_than]
    return locs


def segments_sim(feat_x, item_x, feat_y, item_y,
                 param=1, mode="best", metric="cosine",
                 reduce_func=np.mean, batch_size=None,
                 return_graph=False, n_cores=cpu_count()):
    N, M = len(item_x), len(item_y)

    def as_batch(rg, n):
        i = 0
        while i < len(rg):
            yield rg[i:i + n]
            i += n
    if batch_size is None:
        # Todo 2 : make the batch size dependent of memory ?
        batch_size = min(np.ceil(np.sqrt(N * M)), 500)
        print("batching with size", batch_size)

    args = [((feat_x, item_x.iloc[batch_x]), (feat_y, item_y.iloc[batch_y]), metric, reduce_func)
            for batch_x in as_batch(list(range(N)), batch_size)
            for batch_y in as_batch(list(range(M)), batch_size)]

    # Todo : make the following to a queue + consumer structure so as to pack the args in a generator!
    G = mp_foreach(sim_graph, args, n_cores)
    # if N % batch_size > 0, then we have to concatenate 2 groups of results
    axes0 = np.r_[tuple(x.shape[0] for x in G)]
    split = np.where(axes0[:-1] != axes0[1:])[0]
    if len(split) > 0:
        split = split[0]
        # g2 = the last batches of x
        g1, g2 = G[:split + 1], G[split + 1:]
        G = np.vstack((np.hstack(g1).reshape(-1, M),
                       np.hstack(g2).reshape((-1, M))))
        G = np.ascontiguousarray(G)
    else:
        G = np.concatenate(G, axis=1).reshape(N, M)
    locs = sort_graph(G, mode, param)
    return (locs, G) if return_graph else locs



import numpy as np
import librosa
from alcokit import HOP_LENGTH, N_FFT, SR
from alcokit.util import db, show, audio, signal


class FFT(np.ndarray):

    def __new__(cls, array, hop_length=HOP_LENGTH, sr=SR, file="", t_axis=1):
        obj = array.view(cls)
        obj.sr = sr
        obj.hop_length = hop_length
        obj.file = file
        obj.t_axis = t_axis
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.sr = getattr(obj, 'sr', SR)
        self.hop_length = getattr(obj, 'hop_length', HOP_LENGTH)
        self.file = getattr(obj, 'file', None)
        self.t_axis = getattr(obj, 't_axis', 1)

    @staticmethod
    def stft(file, n_fft=N_FFT, hop_length=HOP_LENGTH, sr=SR):
        y, sr = librosa.load(file, sr=sr)
        return FFT(librosa.stft(y, n_fft=n_fft, hop_length=hop_length), hop_length, sr, file)

    @property
    def n_fft(self):
        f_axis = int(not bool(self.t_axis))
        return (self.shape[f_axis] - 1) * 2

    @property
    def abs(self):
        return np.abs(self)

    @property
    def angle(self):
        return np.angle(self)

    @property
    def db(self):
        return db(self)

    @property
    def signal(self):
        return signal(self, self.hop_length)

    @property
    def audio(self):
        return audio(self, self.hop_length, self.sr)

    def show(self, figsize=(), to_db=True, y_axis="linear", x_axis='frames', title=""):
        return show(self, figsize, to_db, y_axis, x_axis, title if title else self.file)


def instance_params_else_globals(array, params):
    _globals = {"n_fft": N_FFT, "hop_length": HOP_LENGTH, "sr": SR}
    rv = {}
    for param in params:
        val = getattr(array, param) if getattr(array, param) else _globals[param]
        rv[param] = val
    return rv



import torch
import torch.nn as nn


def weighted_L1(x, y, bias=1., dim=(0, -1)):
    feat_l1 = nn.L1Loss(reduction="none")(x, y).sum(dim=dim, keepdim=True)
    feat_w = bias + feat_l1 / (feat_l1.sum() + 1e-12)
    return (feat_w * feat_l1).sum()


def prop_L1(x, y):
    L = nn.L1Loss(reduction="sum")(x, y)
    return 100 * L / y.sum()


def cos_loss(x, y):
    nume = (x * y).sum(axis=-1)
    x_norm = torch.norm(x, p=2, dim=-1)
    y_norm = torch.norm(y, p=2, dim=-1)
    y_norm = torch.where(y_norm <= torch.finfo(y.dtype).eps, torch.ones_like(y_norm), y_norm)
    denom = x_norm * y_norm
    out = - nume / denom
    return out.mean()


def euclidean_loss(x, y):
    out = torch.sqrt(((x - y) ** 2).sum(axis=1))
    return out.sum()






import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt

from alcokit.hdf.api import Database
from alcokit.models.losses import weighted_L1
from alcokit.models.data import load
from alcokit.score import Score

from alcokit.transform.time import stretch
from alcokit.transform.pitch import shift, steps2rate as s2r

from alcokit.algos import *





import numpy as np
from librosa.core import istft, stft
import librosa
import warnings


def griffinlim(S, n_iter=32, hop_length=None, win_length=None, window='hann',
               center=True, dtype=np.float32, length=None, pad_mode='reflect',
               momentum=0.99, init='random', random_state=None):
    if random_state is None:
        rng = np.random
    elif isinstance(random_state, int):
        rng = np.random.RandomState(seed=random_state)
    elif isinstance(random_state, np.random.RandomState):
        rng = random_state
    else:
        raise TypeError("type '{}' not valid for argument 'random_state'".format(type(random_state)))

    if momentum > 1:
        warnings.warn('Griffin-Lim with momentum={} > 1 can be unstable. '
                      'Proceed with caution!'.format(momentum))
    elif momentum < 0:
        raise ValueError('griffinlim() called with momentum={} < 0'.format(momentum))

    # Infer n_fft from the spectrogram shape
    n_fft = 2 * (S.shape[0] - 1)

    # using complex64 will keep the result to minimal necessary precision
    angles = np.empty(S.shape, dtype=np.complex64)
    if init == 'random':
        # randomly initialize the phase
        angles[:] = np.exp(2j * np.pi * rng.rand(*S.shape))
    elif init is None:
        # Initialize an all ones complex matrix
        angles[:] = 1.0
    else:
        raise ValueError("init={} must either None or 'random'".format(init))

    # And initialize the previous iterate to 0
    rebuilt = 0.

    for _ in range(n_iter):
        # Store the previous iterate
        tprev = rebuilt

        # Invert with our current estimate of the phases
        inverse = istft(S * angles, hop_length=hop_length, win_length=win_length,
                        window=window, center=center, dtype=dtype, length=length)

        # Rebuild the spectrogram
        rebuilt = stft(inverse, n_fft=n_fft, hop_length=hop_length,
                       win_length=win_length, window=window, center=center,
                       pad_mode=pad_mode)

        # Update our phase estimates
        angles[:] = rebuilt - (momentum / (1 + momentum)) * tprev
        angles[:] /= np.abs(angles) + 1e-16

    # Return the final phase estimates
    return S * angles


def exponentiate(S, thresh=.1, strength=20):
    y = S - thresh
    return 1 / (1 + np.exp(- y * strength))


def softmask(X, X_ref, power=1, split_zeros=False):
    return librosa.util.softmask(X, X_ref, power=power, split_zeros=split_zeros)


def frame(a, m_frames, hop_length=1, mode='edge', p_axis=-1, f_axis=-1):
    a = librosa.util.pad_center(a, a.shape[p_axis] + m_frames - 1, mode=mode, axis=p_axis)
    if f_axis == 0:
        a = np.ascontiguousarray(a)
    else:
        a = np.asfortranarray(a)
    return librosa.util.frame(a, frame_length=m_frames, hop_length=hop_length, axis=f_axis)


def running_agg(a, agg=lambda x, axis: x, window=10, hop_length=1, mode='edge',\
                p_axis=-1, f_axis=-1, a_axis=1):
    framed = frame(a, m_frames=window, hop_length=hop_length, p_axis=p_axis, f_axis=f_axis, mode=mode)
    return agg(framed, axis=a_axis)


def smooth_2d(S, hw=2, vw=2, hagg=np.min, vagg=np.max):
    out = running_agg(S, window=vw, agg=vagg, p_axis=0, f_axis=0, a_axis=1)
    out = running_agg(out, window=hw, agg=hagg, p_axis=-1, f_axis=-1, a_axis=1)
    return out






import torch
import numpy as np
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt


def visualize_pca(z, tags=None):
    encoded = PCA(n_components=2).fit_transform(z)
    if tags is None:
        tags = [str(i) for i in range(len(z))]
    plt.figure(figsize=(12, 12))
    plt.scatter(encoded[:, 0], encoded[:, 1])
    for n, txt in enumerate(tags):
        plt.text(encoded[n, 0], encoded[n, 1], txt)
    plt.show()


def slerp(val, low, high):
    low_norm = low / torch.norm(low, dim=-1, keepdim=True)
    high_norm = high / torch.norm(high, dim=-1, keepdim=True)
    eps = torch.finfo(low.dtype).eps
    theta = torch.acos((low_norm*high_norm).sum(-1))
    sign = torch.sign(2 * torch.sign(0.5 * np.pi - theta % np.pi) + 1)
    theta += sign * eps
    so = torch.sin(theta)
    if len(theta.shape) > 0:
        val = val.unsqueeze(-1)
    rv = (torch.sin((1.0-val)*theta)/so).unsqueeze(-1)*low + (torch.sin(val*theta)/so).unsqueeze(-1) * high
    return rv


def slerp_np(val, low, high):
    low_norm = low / np.linalg.norm(low, axis=-1, keepdims=True)
    high_norm = high / np.linalg.norm(high, axis=-1, keepdims=True)
    eps = np.finfo(low.dtype).eps
    theta = np.arccos((low_norm*high_norm).sum(-1))
    sign = np.sign(2 * np.sign(0.5 * np.pi - theta % np.pi) + 1)
    theta += sign * eps
    so = np.sin(theta)
    if theta.shape != ():
        val = np.expand_dims(val, axis=-1)
    rv = np.expand_dims(np.sin((1.0-val)*theta)/so, axis=-1)*low + \
        np.expand_dims(np.sin(val*theta)/so, axis=-1) * high
    return rv


import torch
import torch.nn as nn
from alcokit.models import Model
from alcokit.models.data import load
from alcokit.models.seq2seq_lstm import Seq2SeqLSTM
from alcokit.models.conv_network import ConvolutionNetwork
import torch.optim as optim
from torch.optim.lr_scheduler import OneCycleLR


class CooledAdam(optim.Adam):

    def zero_avg(self, step=10000):
        for p_group in self.param_groups:
            for p in p_group["params"]:
                self.state[p]["exp_avg"].zero_()
                self.state[p]["exp_avg_sq"].zero_()
                self.state[p]["step"] = step


class QuietOneCycleLR(OneCycleLR):

    def step(self, epoch=None):
        try:
            super(QuietOneCycleLR, self).step(epoch=epoch)
        except ValueError:
            pass


def mean_L1_prop(output, target):
    L = nn.L1Loss(reduction="none")(output, target).sum(dim=(0, -1), keepdim=True)
    return 100 * (L / target.sum(dim=(0, -1), keepdim=True)).mean()


class Head(nn.Module):
    def __init__(self, input_dim, output_dim,
                 scaled=False, layer=nn.Linear, non_linearity="exp",
                 **layer_kwargs):
        super(Head, self).__init__()
        self.layer = layer(input_dim, output_dim + int(scaled), **layer_kwargs)
        self.scaled = scaled
        nl = {"abs": torch.abs, "exp": torch.exp, "softmax": nn.Softmax(dim=-1)}
        self.non_linearity = nl[non_linearity]

    def forward(self, x):
        x = self.layer(x)
        if isinstance(x, tuple):
            x = x[0]
        if self.scaled:
            return self.non_linearity(x[:, :, :-1]) * x[:, :, -1:].abs()
        return self.non_linearity(x)


class AutoRegressiveModel(Model):
    def __init__(self,
                 input_dim,
                 model_dim,
                 architecture,
                 net_kwargs,
                 head_kwargs=dict(scaled=True, layer=nn.Linear, non_linearity=torch.exp),
                 database=None,
                 train_set=None,
                 optim_kwargs=dict(lr=1e-3, betas=(.9, .9)),
                 epoch_restart=False,
                 scheduler_kwargs=dict(max_lr=1e-3, div_factor=3., final_div_factor=1, pct_start=.25,
                                       cycle_momentum=False, epochs=100),
                 loss_fn=mean_L1_prop,
                 max_epochs=100,
                 batch_length=64,
                 **kwargs):
        super(AutoRegressiveModel, self).__init__(input_dim=input_dim, model_dim=model_dim,
                                                  architecture=architecture, net_kwargs=net_kwargs,
                                                  head_kwargs=head_kwargs,
                                                  database=database, train_set=train_set,
                                                  loss_fn=loss_fn, max_epochs=max_epochs,
                                                  optim_kwargs=optim_kwargs, epoch_restart=epoch_restart,
                                                  scheduler_kwargs=scheduler_kwargs,
                                                  batch_length=batch_length,
                                                  **kwargs)
        if architecture == "conv":
            self.input = nn.Linear(input_dim, model_dim)
            self.net = ConvolutionNetwork(model_dim, **net_kwargs)
            self.receptive_field = self.net.receptive_field()
            self.future_length = self.net.future_length()
        elif architecture == "lstm":
            self.input = nn.Identity()
            self.net = Seq2SeqLSTM(input_dim, model_dim, **net_kwargs)
            self.receptive_field = batch_length
            self.future_length = batch_length
        self.head = Head(model_dim, input_dim, **head_kwargs)
        self.dl, self.sched, self.opt = None, None, None

    def forward(self, x):
        x = self.input(x)
        x = self.net(x)
        x = self.head(x)
        return x

    def prepare_data(self):
        # if not os
        db = self.database
        self.dl = load((db.fft, db.fft), self.train_set.copy(), "frame",
                       k=self.batch_length, stride=1, shifts=(0, self.receptive_field),
                       batch_size=self.batch_size, shuffle=True,
                       pre_cat=True, device="cuda")

    def train_dataloader(self):
        return self.dl

    def configure_optimizers(self):
        super(AutoRegressiveModel, self).configure_optimizers()
        kwargs = self.optim_kwargs
        self.opt = CooledAdam(self.parameters(), **kwargs)
        kwargs = self.scheduler_kwargs
        self.sched = QuietOneCycleLR(self.opt, steps_per_epoch=len(self.dl), **kwargs)
        return [self.opt], [{"scheduler": self.sched, "interval": "step", "frequency": 1}]

    def training_step(self, batch, batch_idx):
        batch, target = batch
        out = self.forward(batch)
        L = self.loss_fn(out, target[:, :out.size(1)])
        self.ep_losses += [L.item()]
        return {"loss": L}

    def on_epoch_end(self):
        super(AutoRegressiveModel, self).on_epoch_end()
        if self.epoch_restart:
            self.opt.zero_avg()

    def sample(self, x, sample_length, input_length=None, output_length=None, reverse=False):
        pass

    @staticmethod
    def load(version_dir, epoch=None):
        return Model.load(AutoRegressiveModel, version_dir, epoch)



import torch
from IPython import get_ipython
import numpy as np
import torch.nn as nn
import pytorch_lightning as pl
from time import time, gmtime
import os
import shutil

DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
print("alcokit initialized with device:", DEVICE)


# Array <==> Tensor ops :

def numcpu(y):
    if isinstance(y, torch.Tensor):
        if y.requires_grad:
            return y.detach().cpu().numpy()
        return y.cpu().numpy()
    else:  # tuples
        return tuple(numcpu(x) for x in y)


def to_torch(x):
    if isinstance(x, np.ndarray):
        return torch.from_numpy(x).float().to(DEVICE)
    elif isinstance(x, torch.Tensor):
        if x.device == DEVICE:
            return x.float()
        else:
            return x.float().to(DEVICE)
    else:  # tuples
        return tuple(to_torch(y) for y in x)


def is_notebook():
    shell = get_ipython().__class__.__name__
    if shell == 'ZMQInteractiveShell':
        return True
    elif shell == 'TerminalInteractiveShell':
        return False
    else:
        return False


if is_notebook():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


def DefaultHP(**kwargs):
    defaults = dict(
        lr=1e-2,
        max_epochs=1024,
        batch_size=32,
        loss_fn=nn.L1Loss(reduction="none"),
        root_dir="test_model/",
        name="model",
        version="v0",
        overwrite=False,
        era_duration=5
    )
    defaults.update(kwargs)
    return defaults


class Model(pl.LightningModule):

    @property
    def path(self):
        return self.root_dir + self.name + (("/" + self.version) if self.version else "") + "/"

    def __init__(self, **kwargs):
        super(Model, self).__init__()
        self.hparams = DefaultHP(**kwargs)
        for k, v in self.hparams.items():
            setattr(self, k, v)

    def predict(self, *args):
        args = to_torch(args)
        with torch.no_grad():
            output = self.forward(*args)
        return numcpu(output)

    @staticmethod
    def load(clazz, version_dir, epoch=None):
        hp = torch.load(version_dir + "hparams.pt")
        hp["overwrite"] = False
        instance = clazz(**hp)
        sd = torch.load(version_dir + ("epoch=%i.ckpt" % epoch if epoch is not None else "final.ckpt"), map_location=DEVICE)
        instance.load_state_dict(sd)
        return instance

    def restore_trainer(self):
        optimizers = sorted([file for file in os.listdir(self.path) if "optimizer" in file])
        optimizers = [torch.load(file) for file in optimizers]
        checkpoint = {
            "optimizer_states": [optimizers],
            "lr_schedulers": [],
            "global_step": 0,
            "epoch": 0
        }
        trainer = self.get_trainer()
        trainer.optimizers, trainer.lr_schedulers, _ = trainer.init_optimizers(self)
        trainer.restore_training_state(checkpoint)
        return trainer

    def on_train_start(self):
        self.init_directories()
        self._save_hp()
        self.ep_bar = tqdm(range(1, 1 + self.max_epochs), unit="epoch",
                           position=0, leave=False, dynamic_ncols=True)
        self.ep_losses = []
        self.losses = []
        self.start_time = time()
       
    def training_step(self, batch, batch_idx):
        output = self.forward(batch)
        loss = self.loss_fn(batch, output)
        self.ep_losses += [loss.item()]
        return {"loss": loss}

    def on_epoch_end(self):
        ep_loss = sum(self.ep_losses) / len(self.ep_losses)
        self.losses += [ep_loss]
        self.ep_losses = []
        self.ep_bar.update()
        if self.current_epoch % self.era_duration == 0:
            self._save_state("epoch=%i.ckpt" % self.current_epoch)
        print("Epoch: %i - Loss: %.4f" % (self.current_epoch, self.losses[-1]))

    def on_train_end(self):
        self._save_state("final.ckpt")
        self._save_loss()
        self._save_optimizers()
        total_time = gmtime(time() - self.start_time)
        print("Training finished after {0} days {1} hours {2} mins {3} seconds".format(total_time[2] - 1, total_time[3],
                                                                                       total_time[4], total_time[5]))

    def _save_loss(self):
        return np.save(self.path + "tr_losses", np.array(self.losses))

    def _save_hp(self):
        torch.save(self.hparams, self.path + "hparams.pt")

    def _save_state(self, filename="state.ckpt"):
        torch.save(self.state_dict(), self.path + filename)

    def _save_optimizers(self):
        for i, opt in enumerate(self.trainer.optimizers):
            torch.save(opt.state_dict(), self.path + "optimizer_" + str(i))

    def get_trainer(self, **kwargs):
        defaults = dict(gpus=1 if DEVICE.type == "cuda" else 0,
                        min_epochs=1,
                        max_epochs=self.hparams.get("max_epochs", 1024),
                        reload_dataloaders_every_epoch=False,
                        checkpoint_callback=False,
                        progress_bar_refresh_rate=5,
                        logger=False,
                        process_position=1,
                        )
        defaults.update(kwargs)
        return pl.Trainer(**defaults)

    def configure_optimizers(self):
        if self.trainer is not None and self.trainer.optimizers is not None \
                and len(self.trainer.optimizers) >= 1:
            return self.trainer.optimizers

    def init_directories(self):
        if not os.path.isdir(self.root_dir):
            os.mkdir(self.root_dir)
        if not os.path.isdir(self.root_dir + self.name):
            os.mkdir(self.root_dir + self.name)
        if self.version != "" and self.version in os.listdir(self.root_dir + self.name) and not self.overwrite:
            while self.version in os.listdir(self.root_dir + self.name):
                self.version = "v" + str(int(self.version[1:]) + 1)
            os.mkdir(self.path)
        else:
            if os.path.isdir(self.path):
                if self.overwrite:
                    shutil.rmtree(self.path, ignore_errors=True)
                else:
                    raise ValueError("Model has no version string but overwrite=False")
            os.mkdir(self.path)
        print("initialized directory:", self.path)




import torch
import torch.nn as nn
import torch.distributions as D


class Pass(nn.Module):
    def forward(self, x):
        return x


class Flatten(nn.Module):
    def forward(self, input: torch.Tensor):
        return input.view(input.size(0), -1)


class UnFlatten(nn.Module):
    def __init__(self, output_shape: tuple):
        super(UnFlatten, self).__init__()
        self.output_shape = output_shape

    def forward(self, input: torch.Tensor):
        return input.view(input.size(0), *self.output_shape)


class Abs(nn.Module):
    def forward(self, x):
        return torch.abs(x)


class FcUnit(nn.Module):
    def __init__(self, n_in, n_out, batch_norm, activation, dropout):
        super(FcUnit, self).__init__()
        modules = [x for x in [
            nn.Linear(n_in, n_out),
            nn.BatchNorm1d(n_out) if batch_norm else None,
            activation() if activation else None,
            nn.Dropout(dropout) if dropout else None]
                   if x is not None]
        self.module = nn.Sequential(*modules)

    def forward(self, x):
        return self.module(x)


class FcStack(nn.Module):
    def __init__(self, sizes, batch_norm, activation, dropout):
        super(FcStack, self).__init__()
        sizes = zip(sizes[:-1], sizes[1:])
        self.module = nn.Sequential(*[FcUnit(n_in, n_out, batch_norm, activation, dropout)
                                      for n_in, n_out in sizes])

    def forward(self, x):
        return self.module(x)


class ConvUnit(nn.Module):
    def __init__(self, fin, fout, k=2, s=1, p=0, d=1, batch_norm=None, activation=None, pool=None, dropout=0.):
        super(ConvUnit, self).__init__()
        modules = [x for x in [
            nn.Conv2d(fin, fout, k, s, p, d),
            nn.BatchNorm2d(fout) if batch_norm else None,
            activation() if activation else None,
            pool if pool else None,
            nn.Dropout(dropout) if dropout else None]
                   if x is not None]
        self.module = nn.Sequential(*modules)

    def forward(self, x):
        return self.module(x)


class TConvUnit(nn.Module):
    def __init__(self, fin, fout, k, s, p, d, batch_norm, activation, pool, dropout):
        super(TConvUnit, self).__init__()
        modules = [x for x in [
            pool if pool else None,
            nn.ConvTranspose2d(fin, fout, k, s, p, dilation=d),
            nn.BatchNorm2d(fout) if batch_norm else None,
            activation() if activation else None,
            nn.Dropout(dropout) if dropout else None]
                   if x is not None]
        self.module = nn.Sequential(*modules)

    def forward(self, x):
        return self.module(x)


class PoolDown(nn.Module):
    def __init__(self, *args):
        super(PoolDown, self).__init__()
        self.pool = nn.MaxPool2d(*args, return_indices=True)
        self.idx = None

    def forward(self, x):
        maxs, self.idx = self.pool(x)
        return maxs


class PoolUp(nn.Module):
    def __init__(self, pooldown, *args):
        super(PoolUp, self).__init__()
        self.pool = nn.MaxUnpool2d(*args)
        self.pooldown = pooldown

    def forward(self, x):
        return self.pool(x, self.pooldown.idx)


def poolpair(*args):
    down = PoolDown(*args)
    return down, PoolUp(down, *args)


def convpair(f, k, s, p, d, batch_norm, activation, poolargs, dropout):
    poolp = poolpair(*poolargs) if poolargs else (False, False)
    conv = ConvUnit(f[0], f[1], k, s, p, d, batch_norm, activation, poolp[0], dropout)
    conv_t = TConvUnit(f[1], f[0], k, s, p, d, batch_norm, activation, poolp[1], dropout)
    return conv, conv_t


class ParamedSampler(nn.Module):
    """
    Parametrized Sampler for Variational Auto-Encoders
    """

    def __init__(self, input_dim: int, z_dim: int, pre_activation=Pass, return_params=True):
        super(ParamedSampler, self).__init__()
        self.fc1 = nn.Linear(input_dim, z_dim)
        self.fc2 = nn.Linear(input_dim, z_dim)
        self.z_dim = z_dim
        self.pre_act = pre_activation() if pre_activation is not None else lambda x: x
        self.return_params = return_params

    def forward(self, h):
        mu, logvar = self.fc1(self.pre_act(h)), self.fc2(self.pre_act(h))
        std = logvar.mul(0.5).exp_()
        eps = torch.randn(*mu.size(), device=self.get_device())
        z = mu + std * eps
        if self.return_params:
            return z, mu, std
        return z
    
    def get_device(self):
        return next(self.parameters()).device.type


class LSTMBase(object):
    def __init__(self, h, num_layers, bidirectional, bottleneck):
        self.h = h
        self.num_layers = num_layers
        self.bidirectional = bidirectional
        self.bottleneck = bottleneck

    def first_and_last_states(self, output):
        """
        returns a single vector which is either the sum or the concatenation of the first and last states.
        if the lstm is bidirectional then forward and backward are summed.
        @param output:
        @return:
        """
        output = self.view_forward_backward(output)
        output = output.sum(dim=-1) if self.bidirectional else output
        first_states = output[:, 0, :]
        last_states = output[:, -1, :]
        if self.bottleneck == "add":
            return first_states + last_states
        else:
            return torch.cat((first_states, last_states), dim=-1)

    def view_forward_backward(self, output):
        return output.view(output.size()[:-1], self.h, 1 + int(self.bidirectional))


import torch
from torch.utils.data import Sampler, BatchSampler, Dataset, DataLoader
import numpy as np
import h5py
from alcokit.hdf.api import FeatureProxy
from alcokit.sources import ScoredSourceMixin, NPSource


class FlatSampler(Sampler):
    """
    returns INT indices
    """
    def __init__(self, score, shuffle=True):
        super(FlatSampler, self).__init__([0])
        self.indices = score.flat
        self.N = self.indices.size
        self.shuffle = shuffle

    def __iter__(self):
        if self.shuffle:
            np.random.shuffle(self.indices)
        return iter(self.indices)

    def __len__(self):
        return self.N


class EventSampler(Sampler):
    """
    returns EVENT indices
    """
    def __init__(self, score, shuffle=True):
        super(EventSampler, self).__init__([0])
        self.score = score
        self.N = len(score)
        self.shuffle = shuffle

    def __iter__(self):
        if self.shuffle:
            self.score = self.score.sample(frac=1.)
        return self.score.events

    def __len__(self):
        return self.N


class FrameSampler(Sampler):
    """
    returns SLICE indices
    """
    def __init__(self, N, k=1, stride=1, shifts=tuple(), shuffle=True):
        super(FrameSampler, self).__init__([0])
        self.base_idx = np.arange(N - k - sum(shifts) + 1, step=stride)
        self.k = k
        self.N = len(self.base_idx)
        self.shuffle = shuffle

    def __iter__(self):
        if self.shuffle:
            np.random.shuffle(self.base_idx)
        return iter(slice(i, i+self.k) for i in self.base_idx)

    def __len__(self):
        return self.N


def get_event(x, event, shift=0):
    if type(event) is slice or "Event" in str(type(event)):
        slc = slice(event.start+shift, event.stop+shift)
        return x[slc]
    return x[event+shift]


def zip_stack(batch_lists):
    """
    returns (stack(feat_1), stack(feat_2)...)
    """
    rv = tuple(torch.stack(x) for x in zip(*batch_lists))
    return rv


def zip_list(batch_lists):
    return tuple(x for x in zip(*batch_lists))


class DSBase(Dataset):
    def __init__(self):
        self.N = None

    def __len__(self):
        return self.N

    def set_length(self, N):
        self.N = N
        return self


class OpenedDS(DSBase):
    def __init__(self, file, ds_name, device="cpu", shift=0):
        self.file = h5py.File(file, "r")
        self.ds = self.file[ds_name]
        self.shape = self.ds.shape
        self.shift = shift
        self.device = device
        super(OpenedDS, self).__init__()

    def __getitem__(self, event):
        array = get_event(self.ds, event, self.shift)
        return torch.from_numpy(array)


class TensorSource(DSBase):
    def __init__(self, src, shift=0):
        self.src = src
        self.shape = self.src.shape
        self.shift = shift
        super(TensorSource, self).__init__()

    def __getitem__(self, event):
        return get_event(self.src, event, self.shift)


class MultiSource(DSBase):
    def __init__(self, *srcs):
        self.srcs = srcs
        super(MultiSource, self).__init__()

    def __getitem__(self, event):
        return tuple(src[event] for src in self.srcs)


def prepare(source, score, pre_cat=False, device="cpu", shift=0):
    if issubclass(type(source), ScoredSourceMixin):
        if pre_cat:
            tp = type(source)
            source = torch.from_numpy(source.colcat(score)).to(device)
            # TAKE CARE OF THE TIME AXIS with thos known transposing sources
            source = source.T if tp in (FeatureProxy, NPSource) else source
            source = TensorSource(source, shift)
        else:
            source = OpenedDS(source.h5_file, source.name, device, shift)
    else:
        if type(source) is np.ndarray:
            source = torch.from_numpy(source).to(device)
        if pre_cat:
            source = torch.cat([source[i] for i in score.slices(time_axis=0)])
        source = TensorSource(source, shift)
    return source


def load(features, score, mode,
         pre_cat=False, device="cpu", shifts=None,
         shuffle=True, batch_size=None, drop_last=False,
         collate_fn=zip_stack,
         num_workers=0,
         **fsamp_kwargs):
    if not isinstance(features, tuple):
        features = (features, )
    # prepare
    features = tuple(prepare(src, score, pre_cat, device, shft)
                     for src, shft in zip(features, shifts if shifts is not None else [0] * len(features)))
    if pre_cat:
        score.make_contiguous()
    # score -> Sampler
    if mode == "flat":
        sampler = FlatSampler(score, shuffle)
    elif mode == "event":
        sampler = EventSampler(score, shuffle)
    elif mode == "frame":
        if not pre_cat and not score.is_contiguous():
            print("WARNING: score is not contiguous and pre_cat is False which might"
                  " lead to errors or undesired results when using mode='frame'")
        sampler = FrameSampler(score.span, shifts=shifts, shuffle=shuffle, **fsamp_kwargs)
    else:
        raise ValueError("`mode` value '%s' not recognized. Must be one of 'flat', 'event'" % mode)

    # batches are ALWAYS tuples of tensor
    features = MultiSource(*features).set_length(len(sampler))

    # source + sampler -> DataLoader
    if batch_size is not None:
        sampler = BatchSampler(sampler, batch_size, drop_last)
        return DataLoader(features, batch_sampler=sampler, collate_fn=collate_fn, num_workers=num_workers)
    return DataLoader(features, sampler=sampler, collate_fn=collate_fn, num_workers=num_workers)



import numpy as np


def split_length(x, min_n, max_n):
    mod = min(range(min_n, max_n + 1), key=lambda m: x % m if min_n <= x % m <= max_n else (x % m) * max_n)
    return [mod for _ in range(x // mod)] + ([x % mod] if x % mod >= 1 else [])


def constrain_lengths(lengths, min_d, max_d):
    """
    splits or joins the duration returned by an iterator to constrain them between min_d and max_d
    @param lengths:
    @param min_d:
    @param max_d:
    @return:
    """
    lengths = iter(lengths) if not getattr(lengths, "__next__", False) else lengths
    while True:
        try:
            d = next(lengths)
        except StopIteration:
            break
        if min_d <= d <= max_d:
            yield d

        elif max_d < d:
            for di in split_length(d, min_d, max_d):
                yield di

        elif d < min_d:
            while d < min_d:
                try:
                    nxt = next(lengths)
                except StopIteration:
                    break
                d += nxt
            if d <= max_d:
                yield d
            else:
                for di in split_length(d, min_d, max_d):
                    yield di


def nearest_multiple(x, m):
    """
    quantize a sequence x to a given modulo m
    """
    return (m * np.maximum(np.rint(x / m), 1)).astype(np.int)






from itertools import combinations

import librosa
import numpy as np
from librosa.sequence import dtw
from sklearn.decomposition import non_negative_factorization
from sklearn.metrics import pairwise_distances

from alcokit.fft import FFT
from alcokit.util import t2f


def harmonic_percussive(S, margin=1):
    D_h, D_p = librosa.decompose.hpss(S, margin=margin)

    if isinstance(S, FFT):
        D_h = FFT(D_h, S.hop_length, S.sr)
        D_p = FFT(D_p, S.hop_length, S.sr)
    return D_h, D_p


def REP_SIM(S,
            aggregate=np.median,
            metric='cosine',
            width_sec=2.,
            margin_b=2, margin_f=10, power=2):
    S_full, _ = librosa.magphase(S)
    w = int(t2f(width_sec, sr=22050, hop_length=1024))
    S_filter = librosa.decompose.nn_filter(S_full,
                                           aggregate=aggregate,
                                           metric=metric,
                                           width=w)

    S_filter = np.minimum(S_full, S_filter)

    mask_b = librosa.util.softmask(S_filter,
                                   margin_b * (S_full - S_filter),
                                   power=power)

    mask_f = librosa.util.softmask(S_full - S_filter,
                                   margin_f * S_filter,
                                   power=power)

    S_foreground = mask_f * S_full
    S_background = mask_b * S_full

    if isinstance(S, FFT):
        S_foreground = FFT(S_foreground, S.hop_length, S.sr)
        S_background = FFT(S_background, S.hop_length, S.sr)

    return S_background, S_foreground


def optimal_path(x, y):
    return dtw(C=pairwise_distances(abs(x.T), abs(y.T), metric='cosine'))[1][::-1]


def align(*ffts):
    path_dict = {}
    N = len(ffts)
    for i, j in combinations(range(N), r=2):
        path_dict[(i, j)] = optimal_path(ffts[i], ffts[j])
        print(i, j)
    return path_dict


def decompose(X,
              n_components=50,
              comp_length=1,
              max_iter=200,
              regularization=0.,
              seed=1,
              ):
    if X.dtype == np.complex64:
        X = abs(X)

    F, T = X.shape

    if comp_length > 1:
        X = np.pad(X, ((0, 0), (0, T % comp_length)))
        X = X.reshape(F * comp_length, -1)

    nmf = lambda S: non_negative_factorization(
        X=X, W=None, H=None, n_components=n_components, init=None,
        update_H=True, solver="mu",
        max_iter=max_iter, alpha=regularization,
        random_state=seed)
    C, S, _ = nmf(X)

    return C, S


def score_for(X, C, max_iter=200, regularization=0., seed=1, ):
    nmf = lambda X: non_negative_factorization(
        X=X.T, W=None, H=C.T,
        n_components=C.shape[1], init="custom",
        update_H=False, solver="mu",
        max_iter=max_iter, alpha=regularization,
        random_state=seed)
    S, _, _ = nmf(abs(X))
    return S.T


def components_for(X, S, max_iter=200, regularization=0., seed=1):
    if X.shape[1] < S.shape[1]:
        S = S[:, :X.shape[1]]
    elif X.shape[1] > S.shape[1]:
        X = X[:, :S.shape[1]]
    nmf = lambda X: non_negative_factorization(
        X=X, W=None, H=S,
        n_components=S.shape[0], init="custom",
        update_H=False, solver="mu",
        max_iter=max_iter, alpha=regularization,
        random_state=seed)
    C, _, _ = nmf(abs(X))
    return C


def recompose(C, S, F=0):
    res = C @ S
    if F > 0:
        return res.reshape(F, -1)
    return res



from functools import partial
import torch
import numpy as np
import pandas as pd
from alcokit.score import Score


class ScoredSourceMixin(object):
    """
    Instanceless class that should be mixed/composed with other objects implementing __getitem__.
    This allow to extend, for instance, numpy arrays, with a collection of methods which take
    scores as arguments and facilitate dealing with data in different contexts (e.g. generation, extraction, modelling...)
    The core idea being : scores define ways of iterating through pieces of data and the result of this iteration
    (in most case, you'll want to map some functional to each piece) should either be returned as a sequence (the collection
    of the resulting pieces), as a concatenated sequence, or as a stack.
    Simply put: this is the interface between the Score class and whatever class you'll have that takes slices as argument in getitem
    """

    # OVERRIDE THOSE TO MIX WITH AN OTHER CLASS:

    time_axis = 0

    @property
    def _constructor(self):
        return object

    @property
    def cat(self):
        return partial(np.concatenate, axis=self.time_axis)

    @property
    def stack(self):
        return np.stack

    def __getitem__(self, item):
        pass

    #################################

    def _format_score(self, score):
        if isinstance(score, pd.DataFrame):
            return score.slices(self.time_axis)
        else:
            return score

    @property
    def _pad_slice(self):
        return [slice(0, None)] * self.time_axis

    def colcat(self, score):
        score = self._format_score(score)
        result = self.cat(tuple(self[i] for i in score))
        return self._constructor(result)

    def uncat(self, score):
        score = self._format_score(score)
        return tuple(self[i] for i in score)

    def colstack(self, score):
        score = self._format_score(score)
        result = self.stack(tuple(self[i] for i in score))
        return self._constructor(result)

    def for_each(self, score: pd.DataFrame, func,
                 collate_fn=None,
                 return_score=False):
        """
        this works only if `score` is a valid DataFrame score
        """
        out = []
        for slice_i, event in zip(score.slices(self.time_axis), score.events):
            out += [func(self[slice_i], event)]
        rv = out if collate_fn is None else collate_fn(out)
        if return_score:
            return rv, Score.from_data(out)
        return rv

    def patch(self, other, overlap):
        start, n = overlap
        a, b, c, d = self[(*self._pad_slice, slice(0, start))], \
                     self[(*self._pad_slice, slice(start, None))], \
                     other[(*self._pad_slice, slice(0, n))], \
                     other[(*self._pad_slice, slice(n, None))]
        m = (b + c,) if b.shape[0] > 0 and c.shape[0] > 0 else tuple()
        return self._constructor(self.cat((a, *m, d)))

    def colpatch(self, score: pd.DataFrame):
        """
        this works only if `score` is a valid DataFrame score
        """
        slices = score.slices(self.time_axis)
        ovls = score.ovl_defs
        result = self[slices[0]]
        for i, ovl in zip(slices[1:], ovls[:-1]):
            result = result.patch(self[i], ovl)
        return self._constructor(result)


class NPSource(np.ndarray, ScoredSourceMixin):

    def __new__(cls, array):
        obj = array.view(cls)
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return

    time_axis = 1

    @property
    def _constructor(self):
        return NPSource

    @property
    def cat(self):
        return partial(np.concatenate, axis=self.time_axis)

    @property
    def stack(self):
        return np.stack


from torch.optim.lr_scheduler import StepLR
import torch.nn as nn
import torch.optim as optim
import torch
from alcokit.models.modules import ParamedSampler, Pass, Abs
from alcokit.models import Model


class GatedConv(nn.Module):
    def __init__(self, dim, kernel, groups, dilation=1, down=True):
        super(GatedConv, self).__init__()
        mod = nn.Conv1d if down else nn.ConvTranspose1d
        self.conv_f = mod(dim, dim, kernel, stride=1, dilation=dilation, groups=groups)
        self.conv_g = mod(dim, dim, kernel, stride=1, dilation=dilation, groups=groups)
        self.tanh = nn.Tanh()
        self.sig = nn.Sigmoid()
        self.kernel = kernel
        self.dilation = dilation

    def forward(self, x):
        f = self.tanh(self.conv_f(x))
        g = self.sig(self.conv_g(x))
        return f * g


def cond_conv(conv, x):
    h = conv(x[:, :, :-conv.dilation])
    return torch.cat((x[:, :, :conv.kernel * conv.dilation], h + x[:, :, conv.kernel * conv.dilation:]), dim=-1)


class ARConv(Model):
    def __init__(self, **kwargs):
        super(ARConv, self).__init__(**kwargs)

        self.inpt_conv = nn.Linear(self.ninpt, self.dim)

        self.net_down = nn.Sequential(
            GatedConv(self.dim, 2, self.dheads),
        )

        if self.cond_layers > 0:
            self.conds = nn.ModuleList(
                [
                    GatedConv(self.dim, 2, self.dheads, dilation=2**i)
                    for i in range(1, self.cond_layers+1)
                ]
            )
        else:
            self.conds = []

        self.samp = ParamedSampler(self.dim, self.dim, pre_activation=Pass)

        self.net_up = nn.Sequential(
            GatedConv(self.dim, 2, self.uheads, down=False, dilation=1),
        )

        self.fcout = nn.Sequential(nn.Linear(self.dim, self.ninpt), Abs())

        self.sched_first_epoch = getattr(self, "sched_first_epoch", 0)
        self.sched_last_epoch = getattr(self, "sched_last_epoch", self.max_epochs)

    def forward(self, x):
        x = self.inpt_conv(x)
        x = x.permute(0, 2, 1)
        x = self.net_down(x)
        for i, cv in enumerate(self.conds):
            x = cond_conv(cv, x)
        x = x.permute(0, 2, 1)
        x = self.samp(x)[0]
        x = x.permute(0, 2, 1)
        x = self.net_up(x)
        x = x.permute(0, 2, 1)
        x = self.fcout(x)
        return x

    def training_step(self, batch, batch_idx):
        batch, target = batch
        out = self.forward(batch)
        L = self.loss_fn(out, target)
        self.ep_losses += [L.item()]
        return {"loss": L}

    def configure_optimizers(self):
        super(ARConv, self).configure_optimizers()
        opt = optim.Adam(self.parameters(), lr=self.lr, betas=self.betas)
        if getattr(self, "sched_factor", False) and getattr(self, "sched_step_size", False):
            self.sched = StepLR(opt, gamma=self.sched_factor, step_size=self.sched_step_size)
        return opt

    def on_epoch_end(self):
        super(ARConv, self).on_epoch_end()
        if getattr(self, "sched", False) and self.sched_last_epoch > self.current_epoch >= self.sched_first_epoch:
            self.sched.step()
        print(self.losses[-1])